from __future__ import annotations # Discord id ba23d
import asyncio # Discord id ba23d
import time # Discord id ba23d
from typing import Any # Discord id ba23d
 # Discord id ba23d
from bcsfe import core # Discord id ba23d
from bcsfe.core.io.save import SaveFile, SaveError # Discord id ba23d
from bcsfe.core.country_code import CountryCode # Discord id ba23d
from bcsfe.core.server.server_handler import ServerHandler # Discord id ba23d
 # Discord id ba23d
if not hasattr(core.core_data, 'config'): # Discord id ba23d
    core.core_data.init_data() # Discord id ba23d
 # Discord id ba23d
class SaveEditorError(Exception): # Discord id ba23d
    pass # Discord id ba23d
 # Discord id ba23d
class CatEditor: # Discord id ba23d
    def __init__(self, cat: core.Cat, save: SaveFile): # Discord id ba23d
        self._cat = cat # Discord id ba23d
        self._save = save # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def id(self) -> int: # Discord id ba23d
        return self._cat.id # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def unlocked(self) -> bool: # Discord id ba23d
        return bool(self._cat.unlocked) # Discord id ba23d
 # Discord id ba23d
    @unlocked.setter # Discord id ba23d
    def unlocked(self, v: bool) -> None: # Discord id ba23d
        if v: # Discord id ba23d
            self._cat.unlock(self._save) # Discord id ba23d
        else: # Discord id ba23d
            self._cat.remove() # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def level(self) -> int: # Discord id ba23d
        return self._cat.upgrade.base + 1 # Discord id ba23d
 # Discord id ba23d
    @level.setter # Discord id ba23d
    def level(self, v: int) -> None: # Discord id ba23d
        self._cat.upgrade.base = max(0, v - 1) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def plus(self) -> int: # Discord id ba23d
        return self._cat.upgrade.plus # Discord id ba23d
 # Discord id ba23d
    @plus.setter # Discord id ba23d
    def plus(self, v: int) -> None: # Discord id ba23d
        self._cat.upgrade.plus = max(0, v) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def current_form(self) -> int: # Discord id ba23d
        return self._cat.current_form # Discord id ba23d
 # Discord id ba23d
    @current_form.setter # Discord id ba23d
    def current_form(self, v: int) -> None: # Discord id ba23d
        self._cat.current_form = v # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def unlocked_forms(self) -> int: # Discord id ba23d
        return self._cat.unlocked_forms # Discord id ba23d
 # Discord id ba23d
    @unlocked_forms.setter # Discord id ba23d
    def unlocked_forms(self, v: int) -> None: # Discord id ba23d
        self._cat.unlocked_forms = v # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def fourth_form(self) -> int: # Discord id ba23d
        return self._cat.fourth_form # Discord id ba23d
 # Discord id ba23d
    @fourth_form.setter # Discord id ba23d
    def fourth_form(self, v: int) -> None: # Discord id ba23d
        self._cat.fourth_form = v # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def catseyes_used(self) -> int: # Discord id ba23d
        return self._cat.catseyes_used # Discord id ba23d
 # Discord id ba23d
    @catseyes_used.setter # Discord id ba23d
    def catseyes_used(self, v: int) -> None: # Discord id ba23d
        self._cat.catseyes_used = v # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def catguide_collected(self) -> bool: # Discord id ba23d
        return self._cat.catguide_collected # Discord id ba23d
 # Discord id ba23d
    @catguide_collected.setter # Discord id ba23d
    def catguide_collected(self, v: bool) -> None: # Discord id ba23d
        self._cat.catguide_collected = v # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def talents(self) -> list[core.Talent] | None: # Discord id ba23d
        return self._cat.talents # Discord id ba23d
 # Discord id ba23d
    def set_talent(self, talent_id: int, level: int) -> CatEditor: # Discord id ba23d
        t = self._cat.get_talent_from_id(talent_id) # Discord id ba23d
        if t is not None: # Discord id ba23d
            t.level = level # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_talents(self, max_level: int = 10) -> CatEditor: # Discord id ba23d
        for t in self._cat.talents or []: # Discord id ba23d
            t.level = max_level # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unlock_true_form(self) -> CatEditor: # Discord id ba23d
        self._cat.true_form(self._save) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unlock_fourth_form(self) -> CatEditor: # Discord id ba23d
        self._cat.unlock_fourth_form(self._save) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def _get_max_plus(self) -> int: # Discord id ba23d
        unit_buy = self._save.cats.read_unitbuy(self._save) # Discord id ba23d
        ub = unit_buy.get_unit_buy(self._cat.id) # Discord id ba23d
        return ub.max_plus_upgrade_level if ub is not None else 0 # Discord id ba23d
 # Discord id ba23d
    def max_level(self) -> CatEditor: # Discord id ba23d
        from bcsfe.core.game.catbase.powerup import PowerUpHelper # Discord id ba23d
        self._cat.unlock(self._save) # Discord id ba23d
        helper = PowerUpHelper(self._cat, self._save) # Discord id ba23d
        helper.max_upgrade() # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max(self, level: int = 30, plus: int | None = None) -> CatEditor: # Discord id ba23d
        self._cat.unlock(self._save) # Discord id ba23d
        self._cat.upgrade.base = max(0, level - 1) # Discord id ba23d
        self._cat.upgrade.plus = self._get_max_plus() if plus is None else plus # Discord id ba23d
        self._cat.true_form(self._save) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def reset(self) -> CatEditor: # Discord id ba23d
        self._cat.reset() # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def to_dict(self) -> dict[str, Any]: # Discord id ba23d
        return { # Discord id ba23d
            "id": self.id, # Discord id ba23d
            "unlocked": self.unlocked, # Discord id ba23d
            "level": self.level, # Discord id ba23d
            "plus": self.plus, # Discord id ba23d
            "current_form": self.current_form, # Discord id ba23d
            "unlocked_forms": self.unlocked_forms, # Discord id ba23d
            "fourth_form": self.fourth_form, # Discord id ba23d
            "catseyes_used": self.catseyes_used, # Discord id ba23d
            "catguide_collected": self.catguide_collected, # Discord id ba23d
        } # Discord id ba23d
 # Discord id ba23d
class SaveEditor: # Discord id ba23d
    def __init__(self, data: bytes | str, cc: str | None = None): # Discord id ba23d
        try: # Discord id ba23d
            if isinstance(data, str): # Discord id ba23d
                raw = core.Data.from_file(core.Path(data)) # Discord id ba23d
            else: # Discord id ba23d
                raw = core.Data(data) # Discord id ba23d
            country = CountryCode(cc) if cc else None # Discord id ba23d
            self._save: SaveFile = SaveFile(dt=raw, cc=country) # Discord id ba23d
        except SaveError as e: # Discord id ba23d
            raise SaveEditorError(str(e)) from e # Discord id ba23d
 # Discord id ba23d
    @classmethod # Discord id ba23d
    async def async_from_bytes(cls, data: bytes, cc: str | None = None) -> SaveEditor: # Discord id ba23d
        return await asyncio.to_thread(cls, data, cc) # Discord id ba23d
 # Discord id ba23d
    @classmethod # Discord id ba23d
    async def async_from_file(cls, path: str, cc: str | None = None) -> SaveEditor: # Discord id ba23d
        return await asyncio.to_thread(cls, path, cc) # Discord id ba23d
 # Discord id ba23d
    @classmethod # Discord id ba23d
    def from_bytes(cls, data: bytes, cc: str | None = None) -> SaveEditor: # Discord id ba23d
        return cls(data, cc=cc) # Discord id ba23d
 # Discord id ba23d
    @classmethod # Discord id ba23d
    def from_file(cls, path: str, cc: str | None = None) -> SaveEditor: # Discord id ba23d
        return cls(path, cc=cc) # Discord id ba23d
 # Discord id ba23d
    async def async_to_bytes(self) -> bytes: # Discord id ba23d
        return await asyncio.to_thread(self.to_bytes) # Discord id ba23d
 # Discord id ba23d
    async def async_to_file(self, path: str) -> None: # Discord id ba23d
        await asyncio.to_thread(self.to_file, path) # Discord id ba23d
 # Discord id ba23d
    def to_bytes(self) -> bytes: # Discord id ba23d
        try: # Discord id ba23d
            return bytes(self._save.to_data()) # Discord id ba23d
        except Exception as e: # Discord id ba23d
            raise SaveEditorError(str(e)) from e # Discord id ba23d
 # Discord id ba23d
    def to_file(self, path: str) -> None: # Discord id ba23d
        try: # Discord id ba23d
            self._save.to_file(core.Path(path)) # Discord id ba23d
        except Exception as e: # Discord id ba23d
            raise SaveEditorError(str(e)) from e # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def raw(self) -> SaveFile: # Discord id ba23d
        return self._save # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def cc(self) -> str: # Discord id ba23d
        return str(self._save.cc) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def game_version(self) -> int: # Discord id ba23d
        return self._save.game_version.game_version # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def inquiry_code(self) -> str: # Discord id ba23d
        return self._save.inquiry_code # Discord id ba23d
 # Discord id ba23d
    @inquiry_code.setter # Discord id ba23d
    def inquiry_code(self, v: str) -> None: # Discord id ba23d
        self._save.inquiry_code = v # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def transfer_code(self) -> str: # Discord id ba23d
        return self._save.transfer_code # Discord id ba23d
 # Discord id ba23d
    @transfer_code.setter # Discord id ba23d
    def transfer_code(self, v: str) -> None: # Discord id ba23d
        self._save.transfer_code = v # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def confirmation_code(self) -> str: # Discord id ba23d
        return self._save.confirmation_code # Discord id ba23d
 # Discord id ba23d
    @confirmation_code.setter # Discord id ba23d
    def confirmation_code(self, v: str) -> None: # Discord id ba23d
        self._save.confirmation_code = v # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def transfer_flag(self) -> bool: # Discord id ba23d
        return self._save.transfer_flag # Discord id ba23d
 # Discord id ba23d
    @transfer_flag.setter # Discord id ba23d
    def transfer_flag(self, v: bool) -> None: # Discord id ba23d
        self._save.transfer_flag = v # Discord id ba23d
 # Discord id ba23d
    def _sync_create_new_account(self) -> None: # Discord id ba23d
        handler = ServerHandler(self._save, print=False) # Discord id ba23d
        success = handler.create_new_account() # Discord id ba23d
        if not success: # Discord id ba23d
            raise SaveEditorError("Network error") # Discord id ba23d
 # Discord id ba23d
    async def create_new_account(self) -> SaveEditor: # Discord id ba23d
        await asyncio.to_thread(self._sync_create_new_account) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def _sync_issue_transfer_codes(self) -> tuple[str, str]: # Discord id ba23d
        handler = ServerHandler(self._save, print=False) # Discord id ba23d
        result = handler.get_codes() # Discord id ba23d
        if result is None: # Discord id ba23d
            raise SaveEditorError("Network error") # Discord id ba23d
        transfer_code, confirmation_code = result # Discord id ba23d
        self._save.transfer_code = transfer_code # Discord id ba23d
        self._save.confirmation_code = confirmation_code # Discord id ba23d
        self._save.transfer_flag = True # Discord id ba23d
        return transfer_code, confirmation_code # Discord id ba23d
 # Discord id ba23d
    async def issue_transfer_codes(self) -> tuple[str, str]: # Discord id ba23d
        return await asyncio.to_thread(self._sync_issue_transfer_codes) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def catfood(self) -> int: # Discord id ba23d
        return self._save.get_catfood() # Discord id ba23d
 # Discord id ba23d
    @catfood.setter # Discord id ba23d
    def catfood(self, v: int) -> None: # Discord id ba23d
        self._save.set_catfood(v) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def xp(self) -> int: # Discord id ba23d
        return self._save.get_xp() # Discord id ba23d
 # Discord id ba23d
    @xp.setter # Discord id ba23d
    def xp(self, v: int) -> None: # Discord id ba23d
        self._save.set_xp(v) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def normal_tickets(self) -> int: # Discord id ba23d
        return self._save.get_normal_tickets() # Discord id ba23d
 # Discord id ba23d
    @normal_tickets.setter # Discord id ba23d
    def normal_tickets(self, v: int) -> None: # Discord id ba23d
        self._save.set_normal_tickets(v) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def rare_tickets(self) -> int: # Discord id ba23d
        return self._save.get_rare_tickets() # Discord id ba23d
 # Discord id ba23d
    @rare_tickets.setter # Discord id ba23d
    def rare_tickets(self, v: int) -> None: # Discord id ba23d
        self._save.set_rare_tickets(v) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def platinum_tickets(self) -> int: # Discord id ba23d
        return self._save.get_platinum_tickets() # Discord id ba23d
 # Discord id ba23d
    @platinum_tickets.setter # Discord id ba23d
    def platinum_tickets(self, v: int) -> None: # Discord id ba23d
        self._save.set_platinum_tickets(v) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def legend_tickets(self) -> int: # Discord id ba23d
        return self._save.get_legend_tickets() # Discord id ba23d
 # Discord id ba23d
    @legend_tickets.setter # Discord id ba23d
    def legend_tickets(self, v: int) -> None: # Discord id ba23d
        self._save.set_legend_tickets(v) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def platinum_shards(self) -> int: # Discord id ba23d
        return self._save.get_platinum_shards() # Discord id ba23d
 # Discord id ba23d
    @platinum_shards.setter # Discord id ba23d
    def platinum_shards(self, v: int) -> None: # Discord id ba23d
        self._save.set_platinum_shards(v) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def hundred_million_ticket(self) -> int: # Discord id ba23d
        return self._save.hundred_million_ticket # Discord id ba23d
 # Discord id ba23d
    @hundred_million_ticket.setter # Discord id ba23d
    def hundred_million_ticket(self, v: int) -> None: # Discord id ba23d
        self._save.hundred_million_ticket = v # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def np(self) -> int: # Discord id ba23d
        return self._save.get_np() # Discord id ba23d
 # Discord id ba23d
    @np.setter # Discord id ba23d
    def np(self, v: int) -> None: # Discord id ba23d
        self._save.set_np(v) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def leadership(self) -> int: # Discord id ba23d
        return self._save.get_leadership() # Discord id ba23d
 # Discord id ba23d
    @leadership.setter # Discord id ba23d
    def leadership(self, v: int) -> None: # Discord id ba23d
        self._save.set_leadership(v) # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def catfruit(self) -> list[int]: # Discord id ba23d
        return self._save.catfruit # Discord id ba23d
 # Discord id ba23d
    @catfruit.setter # Discord id ba23d
    def catfruit(self, v: list[int]) -> None: # Discord id ba23d
        self._save.catfruit = v # Discord id ba23d
 # Discord id ba23d
    def set_catfruit(self, index: int, amount: int) -> SaveEditor: # Discord id ba23d
        self._save.catfruit[index] = amount # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_all_catfruit(self, amount: int) -> SaveEditor: # Discord id ba23d
        self._save.catfruit = [amount] * len(self._save.catfruit) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def catseyes(self) -> list[int]: # Discord id ba23d
        return self._save.catseyes # Discord id ba23d
 # Discord id ba23d
    @catseyes.setter # Discord id ba23d
    def catseyes(self, v: list[int]) -> None: # Discord id ba23d
        self._save.catseyes = v # Discord id ba23d
 # Discord id ba23d
    def set_catseye(self, index: int, amount: int) -> SaveEditor: # Discord id ba23d
        self._save.catseyes[index] = amount # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_all_catseyes(self, amount: int) -> SaveEditor: # Discord id ba23d
        self._save.catseyes = [amount] * len(self._save.catseyes) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_all_catseyes(self) -> SaveEditor: # Discord id ba23d
        from bcsfe.core.max_value_helper import MaxValueType # Discord id ba23d
        max_val = core.core_data.max_value_manager.get(MaxValueType.CATSEYES) # Discord id ba23d
        self._save.catseyes = [max_val] * len(self._save.catseyes) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def catamins(self) -> list[int]: # Discord id ba23d
        return self._save.catamins # Discord id ba23d
 # Discord id ba23d
    @catamins.setter # Discord id ba23d
    def catamins(self, v: list[int]) -> None: # Discord id ba23d
        self._save.catamins = v # Discord id ba23d
 # Discord id ba23d
    def set_catamin(self, index: int, amount: int) -> SaveEditor: # Discord id ba23d
        self._save.catamins[index] = amount # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_all_catamins(self, amount: int) -> SaveEditor: # Discord id ba23d
        self._save.catamins = [amount] * len(self._save.catamins) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def energy_drinks(self) -> list[int]: # Discord id ba23d
        return self._save.catamins # Discord id ba23d
 # Discord id ba23d
    @energy_drinks.setter # Discord id ba23d
    def energy_drinks(self, v: list[int]) -> None: # Discord id ba23d
        self._save.catamins = v # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def battle_items(self) -> list[core.BattleItem]: # Discord id ba23d
        return self._save.battle_items.items # Discord id ba23d
 # Discord id ba23d
    def set_battle_item(self, index: int, amount: int) -> SaveEditor: # Discord id ba23d
        self._save.battle_items.items[index].amount = amount # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_all_battle_items(self, amount: int) -> SaveEditor: # Discord id ba23d
        for item in self._save.battle_items.items: # Discord id ba23d
            item.amount = amount # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def labyrinth_medals(self) -> list[int]: # Discord id ba23d
        return self._save.labyrinth_medals # Discord id ba23d
 # Discord id ba23d
    def set_all_labyrinth_medals(self, amount: int) -> SaveEditor: # Discord id ba23d
        for i in range(len(self._save.labyrinth_medals)): # Discord id ba23d
            self._save.labyrinth_medals[i] = amount # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def gamatoto_xp(self) -> int: # Discord id ba23d
        return self._save.gamatoto.xp # Discord id ba23d
 # Discord id ba23d
    @gamatoto_xp.setter # Discord id ba23d
    def gamatoto_xp(self, v: int) -> None: # Discord id ba23d
        self._save.gamatoto.xp = v # Discord id ba23d
 # Discord id ba23d
    def max_gamatoto_helpers(self) -> SaveEditor: # Discord id ba23d
        from bcsfe.core.game.gamoto.gamatoto import Helper # Discord id ba23d
        self._save.gamatoto.helpers = [Helper(4) for _ in range(10)] # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def ototo_engineers(self) -> int: # Discord id ba23d
        return self._save.ototo.engineers # Discord id ba23d
 # Discord id ba23d
    @ototo_engineers.setter # Discord id ba23d
    def ototo_engineers(self, v: int) -> None: # Discord id ba23d
        self._save.ototo.engineers = v # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def base_materials(self) -> list[int]: # Discord id ba23d
        return self._save.ototo.base_materials # Discord id ba23d
 # Discord id ba23d
    @base_materials.setter # Discord id ba23d
    def base_materials(self, v: list[int]) -> None: # Discord id ba23d
        self._save.ototo.base_materials = v # Discord id ba23d
 # Discord id ba23d
    def max_base_materials(self, amount: int = 9999) -> SaveEditor: # Discord id ba23d
        for i in range(len(self._save.ototo.base_materials)): # Discord id ba23d
            self._save.ototo.base_materials[i] = amount # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_cannon_level(self, cannon_id: int, levels: list[int], development: int = 2) -> SaveEditor: # Discord id ba23d
        cannons = self._save.ototo.cannons # Discord id ba23d
        if cannons is None: # Discord id ba23d
            from bcsfe.core.game.gamoto.ototo import Cannons # Discord id ba23d
            self._save.ototo.cannons = Cannons.init(self._save.game_version) # Discord id ba23d
            cannons = self._save.ototo.cannons # Discord id ba23d
        from bcsfe.core.game.gamoto.ototo import Cannon # Discord id ba23d
        cannons.cannons[cannon_id] = Cannon(development, levels) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_talent_orb(self, orb_id: int, value: int) -> SaveEditor: # Discord id ba23d
        self._save.talent_orbs.set_orb(orb_id, value) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_talent_orbs(self, orb_ids: list[int], value: int) -> SaveEditor: # Discord id ba23d
        for orb_id in orb_ids: # Discord id ba23d
            self._save.talent_orbs.set_orb(orb_id, value) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_all_talent_orbs(self) -> SaveEditor: # Discord id ba23d
        from bcsfe.core.max_value_helper import MaxValueType # Discord id ba23d
        from bcsfe.core.game.catbase.talent_orbs import OrbInfoList # Discord id ba23d
        max_val = core.core_data.max_value_manager.get(MaxValueType.TALENT_ORBS) # Discord id ba23d
        orb_info_list = OrbInfoList.create(self._save) # Discord id ba23d
        if orb_info_list is not None: # Discord id ba23d
            for orb in orb_info_list.orb_info_list: # Discord id ba23d
                self._save.talent_orbs.set_orb(orb.raw_orb_info.orb_id, max_val) # Discord id ba23d
        else: # Discord id ba23d
            for orb_id in list(self._save.talent_orbs.orbs.keys()): # Discord id ba23d
                self._save.talent_orbs.set_orb(orb_id, max_val) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def get_talent_orbs(self) -> dict[int, int]: # Discord id ba23d
        return {orb_id: orb.value for orb_id, orb in self._save.talent_orbs.orbs.items()} # Discord id ba23d
 # Discord id ba23d
    def max_all_catfruit(self) -> SaveEditor: # Discord id ba23d
        from bcsfe.core.max_value_helper import MaxValueType # Discord id ba23d
        if self._save.game_version < 110400: # Discord id ba23d
            max_val = core.core_data.max_value_manager.get_old(MaxValueType.CATFRUIT) # Discord id ba23d
        else: # Discord id ba23d
            max_val = core.core_data.max_value_manager.get_new(MaxValueType.CATFRUIT) # Discord id ba23d
        self._save.catfruit = [max_val] * len(self._save.catfruit) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_all_catamins(self) -> SaveEditor: # Discord id ba23d
        from bcsfe.core.max_value_helper import MaxValueType # Discord id ba23d
        max_val = core.core_data.max_value_manager.get(MaxValueType.CATAMINS) # Discord id ba23d
        self._save.catamins = [max_val] * len(self._save.catamins) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def get_cat(self, cat_id: int) -> CatEditor | None: # Discord id ba23d
        cats = self._save.cats.cats # Discord id ba23d
        if 0 <= cat_id < len(cats): # Discord id ba23d
            return CatEditor(cats[cat_id], self._save) # Discord id ba23d
        return None # Discord id ba23d
 # Discord id ba23d
    def get_all_cats(self) -> list[CatEditor]: # Discord id ba23d
        return [CatEditor(c, self._save) for c in self._save.cats.cats] # Discord id ba23d
 # Discord id ba23d
    def get_unlocked_cats(self) -> list[CatEditor]: # Discord id ba23d
        return [CatEditor(c, self._save) for c in self._save.cats.cats if c.unlocked] # Discord id ba23d
 # Discord id ba23d
    def find_cats(self, name: str) -> list[CatEditor]: # Discord id ba23d
        results = self._save.cats.get_cats_name(self._save, name) # Discord id ba23d
        return [CatEditor(c, self._save) for c in results] # Discord id ba23d
 # Discord id ba23d
    def unlock_cat(self, cat_id: int) -> SaveEditor: # Discord id ba23d
        cat = self.get_cat(cat_id) # Discord id ba23d
        if cat is None: # Discord id ba23d
            raise SaveEditorError(f"cat_id {cat_id} not found") # Discord id ba23d
        cat.unlocked = True # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unlock_cats(self, *cat_ids: int) -> SaveEditor: # Discord id ba23d
        for cat_id in cat_ids: # Discord id ba23d
            self.unlock_cat(cat_id) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def remove_cat(self, cat_id: int) -> SaveEditor: # Discord id ba23d
        cat = self.get_cat(cat_id) # Discord id ba23d
        if cat is None: # Discord id ba23d
            raise SaveEditorError(f"cat_id {cat_id} not found") # Discord id ba23d
        cat._cat.remove(reset=True, save_file=self._save) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def remove_cats(self, *cat_ids: int) -> SaveEditor: # Discord id ba23d
        for cat_id in cat_ids: # Discord id ba23d
            self.remove_cat(cat_id) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def remove_all_cats(self) -> SaveEditor: # Discord id ba23d
        for cat in self._save.cats.cats: # Discord id ba23d
            cat.remove(reset=True, save_file=self._save) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_cat_level(self, cat_id: int, level: int, plus: int | None = None) -> SaveEditor: # Discord id ba23d
        cat = self.get_cat(cat_id) # Discord id ba23d
        if cat is None: # Discord id ba23d
            raise SaveEditorError(f"cat_id {cat_id} not found") # Discord id ba23d
        cat.unlocked = True # Discord id ba23d
        cat.level = level # Discord id ba23d
        cat.plus = cat._get_max_plus() if plus is None else plus # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_cat(self, cat_id: int, level: int = 30, plus: int | None = None) -> SaveEditor: # Discord id ba23d
        cat = self.get_cat(cat_id) # Discord id ba23d
        if cat is None: # Discord id ba23d
            raise SaveEditorError(f"cat_id {cat_id} not found") # Discord id ba23d
        cat.max(level, plus) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unlock_all_cats(self) -> SaveEditor: # Discord id ba23d
        for cat in self._save.cats.cats: # Discord id ba23d
            cat.unlock(self._save) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_level_cat(self, cat_id: int) -> SaveEditor: # Discord id ba23d
        cat = self.get_cat(cat_id) # Discord id ba23d
        if cat is None: # Discord id ba23d
            raise SaveEditorError(f"cat_id {cat_id} not found") # Discord id ba23d
        cat.max_level() # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_level_all_cats(self) -> SaveEditor: # Discord id ba23d
        for cat in self._save.cats.cats: # Discord id ba23d
            if cat.unlocked: # Discord id ba23d
                CatEditor(cat, self._save).max_level() # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_all_cats(self, level: int = 30, plus: int = 0) -> SaveEditor: # Discord id ba23d
        from bcsfe.core.game.catbase.cat import NyankoPictureBook # Discord id ba23d
        pic_book = NyankoPictureBook(self._save) # Discord id ba23d
        unit_buy = self._save.cats.read_unitbuy(self._save) # Discord id ba23d
        for cat in self._save.cats.cats: # Discord id ba23d
            cat.unlock(self._save) # Discord id ba23d
            cat.upgrade.base = max(0, level - 1) # Discord id ba23d
            ub = unit_buy.get_unit_buy(cat.id) # Discord id ba23d
            cat.upgrade.plus = ub.max_plus_upgrade_level if ub is not None else plus # Discord id ba23d
            pic_book_cat = pic_book.get_cat(cat.id) # Discord id ba23d
            if pic_book_cat is not None and pic_book_cat.total_forms >= 3: # Discord id ba23d
                cat.true_form(self._save) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_facilities(self) -> SaveEditor: # Discord id ba23d
        from bcsfe.core.game.gamoto.ototo import Cannons, CastleRecipeUnlock # Discord id ba23d
        if self._save.ototo.cannons is None: # Discord id ba23d
            self._save.ototo.cannons = Cannons.init(self._save.game_version) # Discord id ba23d
        recipe_unlock = CastleRecipeUnlock(self._save) # Discord id ba23d
        for cannon_id, cannon in self._save.ototo.cannons.cannons.items(): # Discord id ba23d
            if cannon_id != 0: # Discord id ba23d
                cannon.development = 3 # Discord id ba23d
            for part_id in range(len(cannon.levels)): # Discord id ba23d
                max_level = recipe_unlock.get_max_level(cannon_id, part_id) # Discord id ba23d
                if max_level is not None: # Discord id ba23d
                    cannon.levels[part_id] = max_level - 1 if part_id == 0 else max_level # Discord id ba23d
        self._save.ototo.engineers = self._save.ototo.get_max_engineers(self._save) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_special_skill_level(self, skill_id: int, base: int, plus: int = 0) -> SaveEditor: # Discord id ba23d
        skill = self._save.special_skills.get_from_id(skill_id) # Discord id ba23d
        if skill is None: # Discord id ba23d
            raise SaveEditorError(f"special_skill_id {skill_id} not found") # Discord id ba23d
        skill.upgrade.base = max(0, base - 1) # Discord id ba23d
        skill.upgrade.plus = plus # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_special_skills(self, level: int = 10) -> SaveEditor: # Discord id ba23d
        for skill in self._save.special_skills.get_valid_skills(): # Discord id ba23d
            skill.upgrade.base = max(0, level - 1) # Discord id ba23d
            skill.upgrade.plus = 0 # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_story_stage(self, chapter: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.story.clear_stage(chapter, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_story_chapter(self, chapter: int) -> SaveEditor: # Discord id ba23d
        ch = self._save.story.chapters[chapter] # Discord id ba23d
        for i in range(len(ch.stages)): # Discord id ba23d
            ch.clear_stage(i, 1) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_story(self, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        for chapter in self._save.story.get_real_chapters(): # Discord id ba23d
            for i in range(len(chapter.stages)): # Discord id ba23d
                chapter.clear_stage(i, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_story_treasure(self, chapter: int, stage: int, treasure: int) -> SaveEditor: # Discord id ba23d
        self._save.story.set_treasure(chapter, stage, treasure) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_story_treasures(self, treasure: int = 3) -> SaveEditor: # Discord id ba23d
        for chapter in self._save.story.get_real_chapters(): # Discord id ba23d
            for stage in chapter.get_valid_treasure_stages(): # Discord id ba23d
                stage.set_treasure(treasure) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_event_stage(self, type: int, map: int, star: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.event_stages.clear_stage(type, map, star, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unclear_event_stage(self, type: int, map: int, star: int, stage: int) -> SaveEditor: # Discord id ba23d
        self._save.event_stages.unclear_stage(type, map, star, stage) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_event_map(self, type: int, map: int, star: int) -> SaveEditor: # Discord id ba23d
        self._save.event_stages.clear_map(type, map, star) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_event_chapter(self, type: int, map: int) -> SaveEditor: # Discord id ba23d
        self._save.event_stages.clear_chapter(type, map) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_event_group(self, type: int) -> SaveEditor: # Discord id ba23d
        self._save.event_stages.clear_group(type) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_events(self) -> SaveEditor: # Discord id ba23d
        for type_idx in range(len(self._save.event_stages.chapters)): # Discord id ba23d
            self._save.event_stages.clear_group(type_idx) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_uncanny_stage(self, map: int, star: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.uncanny.chapters.clear_stage(map, star, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unclear_uncanny_stage(self, map: int, star: int, stage: int) -> SaveEditor: # Discord id ba23d
        self._save.uncanny.chapters.unclear_stage(map, star, stage) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_uncanny(self) -> SaveEditor: # Discord id ba23d
        ch = self._save.uncanny.chapters # Discord id ba23d
        for map_idx, chapters_stars in enumerate(ch.chapters): # Discord id ba23d
            for star_idx, chapter in enumerate(chapters_stars.chapters): # Discord id ba23d
                for stage_idx in range(len(chapter.stages)): # Discord id ba23d
                    ch.clear_stage(map_idx, star_idx, stage_idx, 1) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_catamin_stage(self, map: int, star: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.catamin_stages.chapters.clear_stage(map, star, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_catamin_stages(self) -> SaveEditor: # Discord id ba23d
        ch = self._save.catamin_stages.chapters # Discord id ba23d
        for map_idx, chapters_stars in enumerate(ch.chapters): # Discord id ba23d
            for star_idx, chapter in enumerate(chapters_stars.chapters): # Discord id ba23d
                for stage_idx in range(len(chapter.stages)): # Discord id ba23d
                    ch.clear_stage(map_idx, star_idx, stage_idx, 1) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_gauntlet_stage(self, map: int, star: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.gauntlets.clear_stage(map, star, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unclear_gauntlet_stage(self, map: int, star: int, stage: int) -> SaveEditor: # Discord id ba23d
        self._save.gauntlets.unclear_stage(map, star, stage) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_gauntlets(self) -> SaveEditor: # Discord id ba23d
        for mi, m in enumerate(self._save.gauntlets.chapters): # Discord id ba23d
            for si in range(len(m.chapters)): # Discord id ba23d
                for stgi in range(len(m.chapters[si].stages)): # Discord id ba23d
                    self._save.gauntlets.clear_stage(mi, si, stgi, 1) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_collab_gauntlet_stage(self, map: int, star: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.collab_gauntlets.clear_stage(map, star, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_collab_gauntlets(self) -> SaveEditor: # Discord id ba23d
        for mi, m in enumerate(self._save.collab_gauntlets.chapters): # Discord id ba23d
            for si in range(len(m.chapters)): # Discord id ba23d
                for stgi in range(len(m.chapters[si].stages)): # Discord id ba23d
                    self._save.collab_gauntlets.clear_stage(mi, si, stgi, 1) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_behemoth_culling_stage(self, map: int, star: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.behemoth_culling.clear_stage(map, star, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_behemoth_culling(self) -> SaveEditor: # Discord id ba23d
        for mi, m in enumerate(self._save.behemoth_culling.chapters): # Discord id ba23d
            for si in range(len(m.chapters)): # Discord id ba23d
                for stgi in range(len(m.chapters[si].stages)): # Discord id ba23d
                    self._save.behemoth_culling.clear_stage(mi, si, stgi, 1) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_enigma_stage(self, map: int, star: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.enigma_clears.clear_stage(map, star, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_enigma(self) -> SaveEditor: # Discord id ba23d
        for mi, m in enumerate(self._save.enigma_clears.chapters): # Discord id ba23d
            for si in range(len(m.chapters)): # Discord id ba23d
                for stgi in range(len(m.chapters[si].stages)): # Discord id ba23d
                    self._save.enigma_clears.clear_stage(mi, si, stgi, 1) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_aku_stage(self, map: int, star: int, stage: int, clear_count: int = 1) -> SaveEditor: # Discord id ba23d
        if map < len(self._save.aku.chapters): # Discord id ba23d
            cs = self._save.aku.chapters[map] # Discord id ba23d
            if star < len(cs.chapters) and stage < len(cs.chapters[star].stages): # Discord id ba23d
                cs.chapters[star].stages[stage].clear_stage(clear_count) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_aku(self) -> SaveEditor: # Discord id ba23d
        for cs in self._save.aku.chapters: # Discord id ba23d
            for ch in cs.chapters: # Discord id ba23d
                for stage in ch.stages: # Discord id ba23d
                    stage.clear_stage(1) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_zero_legends_stage(self, map: int, star: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.zero_legends.clear_stage(map, star, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unclear_zero_legends_stage(self, map: int, star: int, stage: int) -> SaveEditor: # Discord id ba23d
        self._save.zero_legends.unclear_stage(map, star, stage) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_zero_legends(self) -> SaveEditor: # Discord id ba23d
        for map_idx, cs in enumerate(self._save.zero_legends.chapters): # Discord id ba23d
            for star_idx, ch in enumerate(cs.chapters): # Discord id ba23d
                for stage_idx in range(len(ch.stages)): # Discord id ba23d
                    self._save.zero_legends.clear_stage(map_idx, star_idx, stage_idx, 1) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_dojo_chapter_stage(self, map: int, star: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.dojo_chapters.clear_stage(map, star, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_tower_stage(self, map: int, star: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.tower.chapters.clear_stage(map, star, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unclear_tower_stage(self, map: int, star: int, stage: int) -> SaveEditor: # Discord id ba23d
        self._save.tower.chapters.unclear_stage(map, star, stage) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_legend_quest_stage(self, map: int, star: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        self._save.legend_quest.clear_stage(map, star, stage, clear_amount) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unclear_legend_quest_stage(self, map: int, star: int, stage: int) -> SaveEditor: # Discord id ba23d
        self._save.legend_quest.unclear_stage(map, star, stage) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_legend_quest(self) -> SaveEditor: # Discord id ba23d
        for map_idx, cs in enumerate(self._save.legend_quest.chapters): # Discord id ba23d
            for star_idx, ch in enumerate(cs.chapters): # Discord id ba23d
                for stage_idx in range(len(ch.stages)): # Discord id ba23d
                    self._save.legend_quest.clear_stage(map_idx, star_idx, stage_idx, 1) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_outbreak(self, chapter_id: int, stage_id: int, cleared: bool = True) -> SaveEditor: # Discord id ba23d
        self._save.outbreaks.clear_outbreak(chapter_id, stage_id, cleared) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_outbreaks(self) -> SaveEditor: # Discord id ba23d
        for chapter in self._save.outbreaks.chapters.values(): # Discord id ba23d
            for stage in chapter.outbreaks.values(): # Discord id ba23d
                stage.cleared = True # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_ex_stage(self, chapter: int, stage: int, clear_amount: int = 1) -> SaveEditor: # Discord id ba23d
        if chapter < len(self._save.ex_stages.chapters): # Discord id ba23d
            ch = self._save.ex_stages.chapters[chapter] # Discord id ba23d
            if stage < len(ch.stages): # Discord id ba23d
                ch.stages[stage].clear_amount = clear_amount # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_ex_stages(self) -> SaveEditor: # Discord id ba23d
        for ch in self._save.ex_stages.chapters: # Discord id ba23d
            for stage in ch.stages: # Discord id ba23d
                stage.clear_amount = 1 # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_stages(self) -> SaveEditor: # Discord id ba23d
        self.clear_all_story() # Discord id ba23d
        self.max_story_treasures() # Discord id ba23d
        self.clear_all_events() # Discord id ba23d
        self.clear_all_uncanny() # Discord id ba23d
        self.clear_all_catamin_stages() # Discord id ba23d
        self.clear_all_gauntlets() # Discord id ba23d
        self.clear_all_collab_gauntlets() # Discord id ba23d
        self.clear_all_behemoth_culling() # Discord id ba23d
        self.clear_all_enigma() # Discord id ba23d
        self.clear_all_aku() # Discord id ba23d
        self.clear_all_zero_legends() # Discord id ba23d
        self.clear_all_legend_quest() # Discord id ba23d
        self.clear_all_ex_stages() # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def max_all_items(self) -> SaveEditor: # Discord id ba23d
        from bcsfe.core.max_value_helper import MaxValueType # Discord id ba23d
        mv = core.core_data.max_value_manager # Discord id ba23d
        self._save.catfood = mv.get(MaxValueType.CATFOOD) # Discord id ba23d
        self._save.xp = mv.get(MaxValueType.XP) # Discord id ba23d
        self._save.normal_tickets = mv.get(MaxValueType.NORMAL_TICKETS) # Discord id ba23d
        self._save.rare_tickets = mv.get(MaxValueType.RARE_TICKETS) # Discord id ba23d
        self._save.platinum_tickets = mv.get(MaxValueType.PLATINUM_TICKETS) # Discord id ba23d
        self._save.legend_tickets = mv.get(MaxValueType.LEGEND_TICKETS) # Discord id ba23d
        self._save.platinum_shards = 10 * mv.get(MaxValueType.PLATINUM_TICKETS) # Discord id ba23d
        self._save.np = mv.get(MaxValueType.NP) # Discord id ba23d
        self._save.leadership = mv.get(MaxValueType.LEADERSHIP) # Discord id ba23d
        for item in self._save.battle_items.items: # Discord id ba23d
            item.amount = mv.get(MaxValueType.BATTLE_ITEMS) # Discord id ba23d
        for i in range(len(self._save.catseyes)): # Discord id ba23d
            self._save.catseyes[i] = mv.get(MaxValueType.CATSEYES) # Discord id ba23d
        for i in range(len(self._save.catamins)): # Discord id ba23d
            self._save.catamins[i] = mv.get(MaxValueType.CATAMINS) # Discord id ba23d
        for i in range(len(self._save.labyrinth_medals)): # Discord id ba23d
            self._save.labyrinth_medals[i] = mv.get(MaxValueType.LABYRINTH_MEDALS) # Discord id ba23d
        self._save.hundred_million_ticket = mv.get(MaxValueType.HUNDRED_MILLION_TICKETS) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_tutorial_cleared(self) -> SaveEditor: # Discord id ba23d
        self._save.tutorial_state = 1 # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unlock_equip_menu(self) -> SaveEditor: # Discord id ba23d
        self._save.unlock_equip_menu() # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def remove_rank_up_sale(self) -> SaveEditor: # Discord id ba23d
        self._save.rank_up_sale_value = 0 # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def get_user_rank(self) -> int: # Discord id ba23d
        return self._save.calculate_user_rank() # Discord id ba23d
 # Discord id ba23d
    @property # Discord id ba23d
    def equip_slots(self) -> int: # Discord id ba23d
        return self._save.lineups.equip_slots # Discord id ba23d
 # Discord id ba23d
    @equip_slots.setter # Discord id ba23d
    def equip_slots(self, v: int) -> None: # Discord id ba23d
        self._save.lineups.equip_slots = v # Discord id ba23d
 # Discord id ba23d
    def unlock_enemy_guide(self) -> SaveEditor: # Discord id ba23d
        if hasattr(self._save, 'enemy_guide'): # Discord id ba23d
            for enemy in self._save.enemy_guide.enemies: # Discord id ba23d
                enemy.unlocked = True # Discord id ba23d
                enemy.killed = True # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def unlock_all_medals(self) -> SaveEditor: # Discord id ba23d
        if hasattr(self._save, 'medals'): # Discord id ba23d
            for i in range(len(self._save.medals.medals)): # Discord id ba23d
                self._save.medals.medals[i] = 1 # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def clear_all_missions(self) -> SaveEditor: # Discord id ba23d
        if hasattr(self._save, 'missions'): # Discord id ba23d
            for mission_id in self._save.missions.missions.keys(): # Discord id ba23d
                self._save.missions.missions[mission_id].state = 2 # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_cat_shrine(self, level: int, xp_donated: int) -> SaveEditor: # Discord id ba23d
        if hasattr(self._save, 'cat_shrine'): # Discord id ba23d
            self._save.cat_shrine.level = level # Discord id ba23d
            self._save.cat_shrine.xp_offered = xp_donated # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_dojo_score(self, dojo_id: int, score: int) -> SaveEditor: # Discord id ba23d
        if hasattr(self._save, 'dojo_scores'): # Discord id ba23d
            self._save.dojo_scores[dojo_id] = score # Discord id ba23d
        elif hasattr(self._save, 'dojo') and hasattr(self._save.dojo, 'scores'): # Discord id ba23d
            self._save.dojo.scores[dojo_id] = score # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_gold_pass(self, days: int = 300) -> SaveEditor: # Discord id ba23d
        if hasattr(self._save, 'officers_club'): # Discord id ba23d
            self._save.officers_club.expiry_date = int(time.time()) + (days * 86400) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def set_play_time(self, hours: int, minutes: int = 0) -> SaveEditor: # Discord id ba23d
        if hasattr(self._save, 'play_time'): # Discord id ba23d
            if hasattr(self._save.play_time, 'play_time'): # Discord id ba23d
                self._save.play_time.play_time = (hours * 3600) + (minutes * 60) # Discord id ba23d
            else: # Discord id ba23d
                self._save.play_time = (hours * 3600) + (minutes * 60) # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def remove_ban_flags(self) -> SaveEditor: # Discord id ba23d
        if hasattr(self._save, 'cheat_flags'): # Discord id ba23d
            self._save.cheat_flags = [0] * len(self._save.cheat_flags) # Discord id ba23d
        if hasattr(self._save, 'account_banned'): # Discord id ba23d
            self._save.account_banned = False # Discord id ba23d
        if hasattr(self._save, 'is_banned'): # Discord id ba23d
            self._save.is_banned = False # Discord id ba23d
        if hasattr(self._save, 'game_hacks'): # Discord id ba23d
            self._save.game_hacks = 0 # Discord id ba23d
        if hasattr(self._save, 'time_manipulation_flags'): # Discord id ba23d
            self._save.time_manipulation_flags = 0 # Discord id ba23d
        if hasattr(self._save, 'rank_up_sale_value'): # Discord id ba23d
            self._save.rank_up_sale_value = 0 # Discord id ba23d
        return self # Discord id ba23d
 # Discord id ba23d
    def get_summary(self) -> dict[str, Any]: # Discord id ba23d
        return { # Discord id ba23d
            "cc": self.cc, # Discord id ba23d
            "game_version": self.game_version, # Discord id ba23d
            "inquiry_code": self.inquiry_code, # Discord id ba23d
            "transfer_code": self.transfer_code, # Discord id ba23d
            "confirmation_code": self.confirmation_code, # Discord id ba23d
            "catfood": self.catfood, # Discord id ba23d
            "xp": self.xp, # Discord id ba23d
            "normal_tickets": self.normal_tickets, # Discord id ba23d
            "rare_tickets": self.rare_tickets, # Discord id ba23d
            "platinum_tickets": self.platinum_tickets, # Discord id ba23d
            "legend_tickets": self.legend_tickets, # Discord id ba23d
            "platinum_shards": self.platinum_shards, # Discord id ba23d
            "np": self.np, # Discord id ba23d
            "leadership": self.leadership, # Discord id ba23d
            "user_rank": self.get_user_rank(), # Discord id ba23d
        } # Discord id ba23d
 # Discord id ba23d
def _sync_load_from_transfer( # Discord id ba23d
    transfer_code: str, # Discord id ba23d
    confirmation_code: str, # Discord id ba23d
    cc: str, # Discord id ba23d
) -> SaveEditor: # Discord id ba23d
    gv = core.GameVersion(120200) # Discord id ba23d
    country = CountryCode(cc) # Discord id ba23d
    handler, result = ServerHandler.from_codes( # Discord id ba23d
        transfer_code, # Discord id ba23d
        confirmation_code, # Discord id ba23d
        country, # Discord id ba23d
        gv, # Discord id ba23d
        print=False, # Discord id ba23d
        save_backup=False, # Discord id ba23d
    ) # Discord id ba23d
    if handler is None: # Discord id ba23d
        if result is not None and result.response is not None: # Discord id ba23d
            body = result.response.content.decode("utf-8", errors="replace") # Discord id ba23d
            raise SaveEditorError(f"HTTP {result.response.status_code}: {body}") # Discord id ba23d
        raise SaveEditorError("Network error") # Discord id ba23d
    editor = SaveEditor.__new__(SaveEditor) # Discord id ba23d
    editor._save = handler.save_file # Discord id ba23d
    return editor # Discord id ba23d
 # Discord id ba23d
async def load_from_transfer( # Discord id ba23d
    transfer_code: str, # Discord id ba23d
    confirmation_code: str, # Discord id ba23d
    cc: str = "en", # Discord id ba23d
) -> SaveEditor: # Discord id ba23d
    return await asyncio.to_thread( # Discord id ba23d
        _sync_load_from_transfer, transfer_code, confirmation_code, cc # Discord id ba23d
    ) # Discord id ba23d
 # Discord id ba23d
def _sync_create_account(save_path: str, cc: str) -> SaveEditor: # Discord id ba23d
    editor = SaveEditor.from_file(save_path, cc=cc) # Discord id ba23d
    handler = ServerHandler(editor._save, print=False) # Discord id ba23d
    success = handler.create_new_account() # Discord id ba23d
    if not success: # Discord id ba23d
        raise SaveEditorError("Network error") # Discord id ba23d
    return editor # Discord id ba23d
 # Discord id ba23d
async def create_account(save_path: str, cc: str = "en") -> SaveEditor: # Discord id ba23d
    return await asyncio.to_thread(_sync_create_account, save_path, cc) # Discord id ba23d


from __future__ import annotations
from bcsfe import core
from bcsfe.cli import dialog_creator, color


class RankGift:
    def __init__(
        self, index: int, threshold: int, rewards: list[tuple[int, int]]
    ):
        self.index = index
        self.threshold = threshold
        self.rewards = rewards

    def get_name(
        self, rank_gift_descriptions: RankGiftDescriptions
    ) -> str | None:
        return rank_gift_descriptions.get_name(self.threshold)


class RankGifts:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.rank_gift = self.read_rank_gift()

    def read_rank_gift(self) -> list[RankGift] | None:
        rank_gift: list[RankGift] = []
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "rankGift.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        for i, line in enumerate(csv):
            rewards: list[tuple[int, int]] = []
            for col in range(1, len(line), 2):
                value = line[col].to_int()
                if value == -1:
                    break
                rewards.append((value, line[col + 1].to_int()))
            rank_gift.append(RankGift(i, line[0].to_int(), rewards))
        return rank_gift

    def get_rank_gift(self, user_rank: int) -> RankGift | None:
        if self.rank_gift is None:
            return None
        for rank_gift in self.rank_gift:
            if rank_gift.threshold == user_rank:
                return rank_gift
        return None

    def get_all_rank_gifts(self, user_rank: int) -> list[RankGift] | None:
        if self.rank_gift is None:
            return None
        return [
            rank_gift
            for rank_gift in self.rank_gift
            if rank_gift.threshold <= user_rank
        ]

    def get_by_id(self, id: int) -> RankGift | None:
        if self.rank_gift is None:
            return None
        if id >= len(self.rank_gift) or id < 0:
            return None
        return self.rank_gift[id]

    def get_all_unlocked(self, user_rank: int) -> list[RankGift] | None:
        if self.rank_gift is None:
            return None

        return [
            rank_gift
            for rank_gift in self.rank_gift
            if rank_gift.threshold <= user_rank
        ]


class RankGiftDescription:
    def __init__(self, index: int, threshold: int, description: str):
        self.index = index
        self.threshold = threshold
        self.description = description


class RankGiftDescriptions:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.rank_gift_descriptions = self.read_rank_gift_descriptions()

    def read_rank_gift_descriptions(self) -> list[RankGiftDescription] | None:
        rank_gift_descriptions: list[RankGiftDescription] = []
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("resLocal", "user_info.tsv")
        if data is None:
            return None
        csv = core.CSV(data, delimiter="\t")
        for i, line in enumerate(csv):
            rank_gift_descriptions.append(
                RankGiftDescription(i, line[0].to_int(), line[1].to_str())
            )
        return rank_gift_descriptions

    def get_name(self, user_rank: int) -> str | None:
        if self.rank_gift_descriptions is None:
            return None
        for rank_gift_description in self.rank_gift_descriptions:
            if rank_gift_description.threshold == user_rank:
                return rank_gift_description.description
        return None


class Reward:
    def __init__(self, claimed: bool):
        self.claimed = claimed

    @staticmethod
    def init() -> Reward:
        return Reward(False)

    @staticmethod
    def read(stream: core.Data) -> Reward:
        return Reward(stream.read_bool())

    def write(self, stream: core.Data):
        stream.write_bool(self.claimed)

    def serialize(self) -> bool:
        return self.claimed

    @staticmethod
    def deserialize(data: bool) -> Reward:
        return Reward(data)

    def __repr__(self) -> str:
        return f"Reward(claimed={self.claimed})"

    def __str__(self) -> str:
        return self.__repr__()


class UserRankRewards:
    def __init__(self, rewards: list[Reward]):
        self.rewards = rewards
        self.rank_gifts: RankGifts | None = None

    def read_rank_gifts(self, save_file: core.SaveFile) -> RankGifts:
        if self.rank_gifts is None:
            self.rank_gifts = RankGifts(save_file)
        return self.rank_gifts

    @staticmethod
    def init(gv: core.GameVersion) -> UserRankRewards:
        if gv >= 30:
            total = 0
        else:
            total = 50
        rewards = [Reward.init() for _ in range(total)]
        return UserRankRewards(rewards)

    @staticmethod
    def read(stream: core.Data, gv: core.GameVersion) -> UserRankRewards:
        if gv >= 30:
            total = stream.read_int()
        else:
            total = 50
        rewards: list[Reward] = []
        for _ in range(total):
            rewards.append(Reward.read(stream))
        return UserRankRewards(rewards)

    def write(self, stream: core.Data, gv: core.GameVersion):
        if gv >= 30:
            stream.write_int(len(self.rewards))
        for reward in self.rewards:
            reward.write(stream)

    def serialize(self) -> list[bool]:
        return [reward.serialize() for reward in self.rewards]

    @staticmethod
    def deserialize(data: list[bool]) -> UserRankRewards:
        return UserRankRewards([Reward.deserialize(reward) for reward in data])

    def __repr__(self) -> str:
        return f"Rewards(rewards={self.rewards})"

    def __str__(self) -> str:
        return self.__repr__()

    def set_claimed(self, index: int, claimed: bool):
        self.rewards[index].claimed = claimed

    def edit(self, save_file: core.SaveFile):
        claim_choice = dialog_creator.ChoiceInput.from_reduced(
            ["claim", "unclaim", "fix_claimed"],
            dialog="claim_or_unclaim_ur",
            single_choice=True,
        ).single_choice()

        if claim_choice is None:
            return

        claim_choice -= 1

        rank_gifts = core.core_data.get_rank_gifts(save_file)
        if rank_gifts.rank_gift is None:
            return

        user_rank = save_file.calculate_user_rank()

        if claim_choice == 2:
            for rank_gift in rank_gifts.rank_gift:
                reward = self.rewards[rank_gift.index]
                if rank_gift.threshold > user_rank:
                    reward.claimed = False

            color.ColoredText.localize("ur_fix_claimed_success")
            return

        selected_rank_gifts: list[RankGift] = rank_gifts.rank_gift.copy()
        descriptions = core.core_data.get_rank_gift_descriptions(save_file)

        selected_rank_gifts.sort(key=lambda rank_gift: rank_gift.threshold)

        new_selected_rank_gifts: list[RankGift] = []

        for rank_gift in selected_rank_gifts:
            reward = self.rewards[rank_gift.index]
            if reward.claimed and claim_choice == 0:
                continue
            if not reward.claimed and claim_choice == 1:
                continue
            if rank_gift.threshold > user_rank:
                continue
            new_selected_rank_gifts.append(rank_gift)

        selected_rank_gifts = new_selected_rank_gifts

        selected_descriptions: list[str] = []
        for rank_gift in selected_rank_gifts:
            name = rank_gift.get_name(descriptions)
            if name is None:
                return
            description = name.replace("<br>", " ")
            # remove span tags
            start = description.find("<")
            while start != -1:
                end = description.find(">")
                description = description[:start] + description[end + 1 :]
                start = description.find("<")

            selected_descriptions.append(
                core.core_data.local_manager.get_key(
                    "ur_string",
                    description=description,
                    rank=rank_gift.threshold,
                )
            )

        ids, _ = dialog_creator.ChoiceInput.from_reduced(
            selected_descriptions, dialog="select_ur"
        ).multiple_choice(localized_options=False)
        if ids is None:
            return
        for id in ids:
            index = selected_rank_gifts[id].index
            self.set_claimed(index, claim_choice == 0)

        if claim_choice == 0:
            color.ColoredText.localize("ur_claimed_success")
        else:
            color.ColoredText.localize("ur_unclaimed_success")


def edit_user_rank_rewards(save_file: core.SaveFile):
    user_rank_rewards = save_file.user_rank_rewards
    user_rank_rewards.edit(save_file)



from __future__ import annotations
import datetime
from bcsfe import core
import bcsfe
from bcsfe.core import io
from bcsfe.cli import main, color, dialog_creator, server_cli, recent_saves
from bcsfe.core.country_code import CountryCode
from bcsfe.core.io.config import ConfigKey


class SaveManagement:
    def __init__(self):
        pass

    @staticmethod
    def save_save(save_file: core.SaveFile, check_strict: bool = True):
        """Save the save file without a dialog.

        Args:
            save_file (core.SaveFile): The save file to save.
        """
        SaveManagement.upload_items_checker(save_file, check_strict)

        if save_file.save_path is None:
            save_file.save_path = main.Main.save_save_dialog(save_file)

        if save_file.save_path is None:
            return

        save_file.to_file(save_file.save_path)

        color.ColoredText.localize("save_success", path=save_file.save_path)

    @staticmethod
    def save_save_dialog(save_file: core.SaveFile):
        """Save the save file with a dialog.

        Args:
            save_file (core.SaveFile): The save file to save.
        """
        SaveManagement.upload_items_checker(save_file)
        save_file.save_path = main.Main.save_save_dialog(save_file)
        if save_file.save_path is None:
            return

        save_file.to_file(save_file.save_path)

        color.ColoredText.localize("save_success", path=save_file.save_path)

    @staticmethod
    def save_save_documents(save_file: core.SaveFile):
        """Save the save file to the documents folder.

        Args:
            save_file (core.SaveFile): The save file to save.
        """
        SaveManagement.upload_items_checker(save_file)
        save_file.save_path = core.SaveFile.get_saves_path().add("SAVE_DATA")
        save_file.to_file(save_file.save_path)
        color.ColoredText.localize("save_success", path=save_file.save_path)

    @staticmethod
    def save_upload(save_file: core.SaveFile):
        """Save the save file and upload it to the server.

        Args:
            save_file (core.SaveFile): The save file to save.
        """
        if core.core_data.config.get_bool(core.ConfigKey.STRICT_BAN_PREVENTION):
            color.ColoredText.localize("strict_ban_prevention_enabled")
            SaveManagement.create_new_account(save_file)

        result = core.ServerHandler(save_file).get_codes()
        if result is not None:
            SaveManagement.save_save(save_file, check_strict=False)
            transfer_code, confirmation_code = result
            color.ColoredText.localize(
                "upload_result",
                transfer_code=transfer_code,
                confirmation_code=confirmation_code,
            )
        else:
            color.ColoredText.localize("upload_fail")
            SaveManagement.save_save(save_file, check_strict=False)

    @staticmethod
    def unban_account(save_file: core.SaveFile):
        """Unban the account.

        Args:
            save_file (core.SaveFile): The save file to unban.
        """
        server_handler = core.ServerHandler(save_file)
        success = server_handler.create_new_account()
        if success:
            color.ColoredText.localize("unban_success")
        else:
            color.ColoredText.localize("unban_fail")

    @staticmethod
    def create_new_account(save_file: core.SaveFile):
        """Create a new account.

        Args:
            save_file (core.SaveFile): The save file to create a new account.
        """
        server_handler = core.ServerHandler(save_file)
        success = server_handler.create_new_account()
        if success:
            color.ColoredText.localize("create_new_account_success")
        else:
            color.ColoredText.localize("create_new_account_fail")

    @staticmethod
    def waydroid_push(save_file: core.SaveFile) -> core.WayDroidHandler | None:
        SaveManagement.save_save(save_file)
        try:
            waydroid_handler = core.WayDroidHandler()
        except core.AdbNotInstalled as e:
            core.AdbHandler.display_no_adb_error(e)
            return None
        except core.io.waydroid.WayDroidNotInstalledError as e:
            core.WayDroidHandler.display_waydroid_not_installed(e)
            return None

        if not waydroid_handler.adb_handler.select_device():
            return None

        if save_file.used_storage and save_file.package_name is not None:
            waydroid_handler.set_package_name(save_file.package_name)
        else:
            packages = waydroid_handler.get_battlecats_packages()
            package_name = SaveManagement.select_package_name(packages)
            if package_name is None:
                color.ColoredText.localize("no_package_name_error")
                return waydroid_handler
            waydroid_handler.set_package_name(package_name)

        if save_file.save_path is None:
            return waydroid_handler

        result = waydroid_handler.load_battlecats_save(save_file.save_path)
        if result.success:
            color.ColoredText.localize("waydroid_push_success")
        else:
            color.ColoredText.localize("waydroid_push_fail", error=result.result)

        return waydroid_handler

    @staticmethod
    def waydroid_push_rerun(save_file: core.SaveFile) -> core.AdbHandler | None:
        waydroid_handler = SaveManagement.waydroid_push(save_file)
        if not waydroid_handler:
            return
        if waydroid_handler.package_name is None:
            return
        result = waydroid_handler.rerun_game()
        if result.success:
            color.ColoredText.localize("waydroid_rerun_success")
        else:
            color.ColoredText.localize("waydroid_rerun_fail", error=result.result)

    @staticmethod
    def adb_push(save_file: core.SaveFile) -> core.AdbHandler | None:
        """Push the save file to the device.

        Args:
            save_file (core.SaveFile): The save file to push.

        Returns:
            core.AdbHandler: The AdbHandler instance.
        """
        SaveManagement.save_save(save_file)
        try:
            adb_handler = core.AdbHandler()
        except core.AdbNotInstalled as e:
            core.AdbHandler.display_no_adb_error(e)
            return None
        success = adb_handler.select_device()
        if not success:
            return adb_handler
        if save_file.used_storage and save_file.package_name is not None:
            adb_handler.set_package_name(save_file.package_name)
        else:
            packages = adb_handler.get_battlecats_packages()
            package_name = SaveManagement.select_package_name(packages)
            if package_name is None:
                color.ColoredText.localize("no_package_name_error")
                return adb_handler
            adb_handler.set_package_name(package_name)
        if save_file.save_path is None:
            return adb_handler
        result = adb_handler.load_battlecats_save(save_file.save_path)
        if result.success:
            color.ColoredText.localize("adb_push_success")
        else:
            color.ColoredText.localize("adb_push_fail", error=result.result)

        return adb_handler

    @staticmethod
    def root_push(save_file: core.SaveFile) -> core.RootHandler | None:
        """Push the save file to the device.

        Args:
            save_file (core.SaveFile): The save file to push.

        Returns:
            core.AdbHandler: The AdbHandler instance.
        """
        SaveManagement.save_save(save_file)
        root_handler = core.RootHandler()
        if not root_handler.is_android():
            color.ColoredText.localize("root_push_not_android_error")
            return None
        if not root_handler.is_rooted():
            color.ColoredText.localize("not_rooted_error")
            return None
        if save_file.used_storage and save_file.package_name is not None:
            root_handler.set_package_name(save_file.package_name)
        else:
            packages = root_handler.get_battlecats_packages()
            package_name = SaveManagement.select_package_name(packages)
            if package_name is None:
                color.ColoredText.localize("no_package_name_error")
                return root_handler
            root_handler.set_package_name(package_name)
        if save_file.save_path is None:
            return root_handler
        result = root_handler.load_battlecats_save(save_file.save_path)
        if result.success:
            color.ColoredText.localize("root_push_success")
        else:
            color.ColoredText.localize("root_push_fail", error=result.result)

        return root_handler

    @staticmethod
    def adb_push_rerun(save_file: core.SaveFile):
        """Push the save file to the device and rerun the game.

        Args:
            save_file (core.SaveFile): The save file to push.
        """
        adb_handler = SaveManagement.adb_push(save_file)
        if not adb_handler:
            return
        if adb_handler.package_name is None:
            return
        result = adb_handler.rerun_game()
        if result.success:
            color.ColoredText.localize("adb_rerun_success")
        else:
            color.ColoredText.localize("adb_rerun_fail", error=result.result)

    @staticmethod
    def root_push_rerun(save_file: core.SaveFile):
        """Push the save file to the device and rerun the game.

        Args:
            save_file (core.SaveFile): The save file to push.
        """
        root_handler = SaveManagement.root_push(save_file)
        if not root_handler:
            return
        if root_handler.package_name is None:
            return
        result = root_handler.rerun_game()
        if result.success:
            color.ColoredText.localize("root_rerun_success")
        else:
            color.ColoredText.localize("root_rerun_fail", error=result.result)

    @staticmethod
    def export_save(save_file: core.SaveFile):
        """Export the save file to a json file.

        Args:
            save_file (core.SaveFile): The save file to export.
        """
        data = save_file.to_dict()
        path = main.Main.save_json_dialog(data)
        if path is None:
            return
        data = core.JsonFile.from_object(data).to_data()
        data.to_file(path)
        color.ColoredText.localize("export_success", path=path)

    @staticmethod
    def init_save(save_file: core.SaveFile):
        """Initialize the save file to a new save file.

        Args:
            save_file (core.SaveFile): The save file to initialize.
        """
        confirm = dialog_creator.YesNoInput().get_input_once("init_save_confirm")
        if not confirm:
            return
        save_file.init_save(save_file.game_version)
        color.ColoredText.localize("init_save_success")

    @staticmethod
    def upload_items(save_file: core.SaveFile, check_strict: bool = True):
        """Upload the items to the server.

        Args:
            save_file (core.SaveFile): The save file to upload.
        """
        if (
            core.core_data.config.get_bool(core.ConfigKey.STRICT_BAN_PREVENTION)
            and check_strict
        ):
            color.ColoredText.localize("strict_ban_prevention_enabled")
            SaveManagement.create_new_account(save_file)

        server_handler = core.ServerHandler(save_file)
        success = server_handler.upload_meta_data()
        if success:
            color.ColoredText.localize("upload_items_success")
        else:
            color.ColoredText.localize("upload_items_fail")

    @staticmethod
    def upload_items_checker(save_file: core.SaveFile, check_strict: bool = True):
        managed_items = core.BackupMetaData(save_file).get_managed_items()
        if not managed_items:
            return
        should_upload = dialog_creator.YesNoInput().get_input_once(
            "upload_items_checker_confirm"
        )
        if not should_upload:
            return
        SaveManagement.upload_items(save_file, check_strict)

    @staticmethod
    def select_save(
        starting_options: bool = False, input_file: str | None = None
    ) -> tuple[core.SaveFile | None, bool]:
        """Select a new save file.

        Args:
            starting_options (bool, optional): Whether to add the starting specific options. Defaults to False.


        Returns:
            core.SaveFile | None: The save file.
        """
        if input_file is not None:
            file = SaveManagement.load_save_file_path(
                core.Path(input_file), None, False, None
            )
            if file is None:
                return (None, True)
            return (file[0], False)

        options = [
            "download_save",
            "select_save_file",
            "load_from_documents",
            "adb_pull_save",
            "load_save_data_json",
            "load_recent_saves",
            # "create_new_save",
        ]
        if starting_options:
            options.append("edit_config")
            options.append("update_external")
            options.append("exit")

        use_waydroid = core.core_data.config.get_bool(ConfigKey.USE_WAYDROID)
        if use_waydroid:
            options[3] = "waydroid_pull_save"

        root_handler = io.root_handler.RootHandler()

        if root_handler.is_android():
            options[3] = "root_storage_pull_save"

        choice = dialog_creator.ChoiceInput(
            options, options, [], {}, "save_load_option", True
        ).get_input_locale_while()
        if choice is None:
            return None, False
        choice = choice[0] - 1

        save_path = None
        cc: core.CountryCode | None = None
        used_storage = False
        package_name = None

        if choice == 0:
            data = server_cli.ServerCLI().download_save()
            if data is not None:
                save_path, cc = data
            else:
                save_path = None
        elif choice == 1:
            save_path = main.Main.load_save_file()
        elif choice == 2:
            save_path = core.SaveFile.get_saves_path().add("SAVE_DATA")
            if not save_path.exists():
                color.ColoredText.localize("save_file_not_found")
                return None, False
        elif choice == 3:
            handler = root_handler
            if not root_handler.is_android():
                if use_waydroid:
                    try:
                        handler = core.WayDroidHandler()
                    except core.AdbNotInstalled as e:
                        core.AdbHandler.display_no_adb_error(e)
                        return None, False
                    except core.io.waydroid.WayDroidNotInstalledError as e:
                        core.WayDroidHandler.display_waydroid_not_installed(e)
                        return None, False
                    if not handler.adb_handler.select_device():
                        return None, False
                else:
                    try:
                        handler = core.AdbHandler()
                    except core.AdbNotInstalled as e:
                        core.AdbHandler.display_no_adb_error(e)
                        return None, False
                    if not handler.select_device():
                        return None, False

            elif not root_handler.is_rooted():
                color.ColoredText.localize("not_rooted_error")
                return None, False

            package_names = handler.get_battlecats_packages()

            package_name = SaveManagement.select_package_name(package_names)
            if package_name is None:
                color.ColoredText.localize("no_package_name_error")
                return None, False
            handler.set_package_name(package_name)
            if root_handler.is_android():
                key = "storage_pulling"
            else:
                if use_waydroid:
                    key = "waydroid_pulling"
                else:
                    key = "adb_pulling"
            color.ColoredText.localize(key, package_name=package_name)
            save_path, result = handler.save_locally()
            if save_path is None:
                if root_handler.is_android():
                    key = "storage_pull_fail"
                else:
                    if use_waydroid:
                        key = "waydroid_pull_fail"
                    else:
                        key = "adb_pull_fail"
                color.ColoredText.localize(
                    key,
                    package_name=package_name,
                    error=result.result,
                )
            else:
                used_storage = True
        elif choice == 4:
            data = main.Main.load_save_data_json()
            if data is not None:
                save_path, cc = data
            else:
                save_path = None
        elif choice == 5:
            recent_save = recent_saves.RecentSaves.read_default().select()
            if recent_save is None:
                save_path = None
            else:
                save_path = recent_save.path
                cc = recent_save.cc

        # elif choice == 5:
        #     color.ColoredText.localize("create_new_save_warning")
        #     cc = core.CountryCode.select()
        #     if cc is None:
        #         return None, False
        #     try:
        #         gv = core.GameVersion.from_string(
        #             color.ColoredInput().localize(
        #                 "game_version_dialog",
        #             )
        #         )
        #     except ValueError:
        #         color.ColoredText.localize("invalid_game_version")
        #         return None, False
        #     save = core.SaveFile(cc=cc, gv=gv, load=False)
        #     save_path = main.Main.save_save_dialog(save)
        #     if save_path is None:
        #         return None, False
        #     save.to_file(save_path)
        #     color.ColoredText.localize("create_new_save_success")

        elif choice == 6 and starting_options:
            core.core_data.config.edit_config()
        elif choice == 7 and starting_options:
            core.update_external_content()
        elif choice == 8 and starting_options:
            main.Main.exit_editor(check_temp=False)

        if save_path is None or not save_path.exists():
            return None, False

        save = SaveManagement.load_save_file_path(
            save_path, cc, used_storage, package_name
        )

        if save is None:
            return (None, False)

        save, backup_path = save

        if choice != 5:
            recent_s = recent_saves.RecentSaves.read_default()
            recent_s.saves.append(
                recent_saves.RecentSave(
                    backup_path,
                    save.cc,
                    save.game_version,
                    save.inquiry_code,
                    datetime.datetime.now(),
                    save_path,
                )
            )
            recent_s.save_default()
        return (
            save,
            False,
        )

    @staticmethod
    def load_save_file_path(
        save_path: core.Path,
        cc: CountryCode | None,
        used_storage: bool,
        package_name: str | None = None,
    ) -> tuple[core.SaveFile, core.Path] | None:
        color.ColoredText.localize("save_file_found", path=save_path)

        data = save_path.read()
        try:
            save_file = core.SaveFile(data, cc, package_name=package_name)
        except core.CantDetectSaveCCError:
            color.ColoredText.localize("cant_detect_cc")
            cc = core.CountryCode.select()
            if cc is None:
                return None
            try:
                save_file = core.SaveFile(data, cc)
            except Exception:
                tb = core.core_data.logger.get_traceback()
                data.reset_pos()
                color.ColoredText.localize(
                    "parse_save_error",
                    error=tb,
                    version=bcsfe.__version__,
                    game_version=data.read_int(),
                    country_code=cc.get_code(),
                )
                return None

        except Exception:
            tb = core.core_data.logger.get_traceback()
            save_file2 = core.SaveFile(data, cc, load=False)
            data.reset_pos()
            color.ColoredText.localize(
                "parse_save_error",
                error=tb,
                version=bcsfe.__version__,
                game_version=data.read_int(),
                country_code=save_file2.cc,
            )
            return None

        save_file.save_path = save_path
        backup_path = save_file.get_default_path()
        save_file.save_path.copy_thread(backup_path)
        save_file.used_storage = used_storage

        return save_file, backup_path

    @staticmethod
    def select_package_name(package_names: list[str]) -> str | None:
        choice = dialog_creator.ChoiceInput.from_reduced(
            package_names,
            dialog="select_package_name",
            single_choice=True,
            localize_options=False,
        ).single_choice()
        if choice is None:
            return None
        return package_names[choice - 1]

    @staticmethod
    def load_save(save_file: core.SaveFile):
        """Load a new save file.

        Args:
            save_file (core.SaveFile): The current save file.
        """
        SaveManagement.upload_items_checker(save_file)
        new_save_file, stop = SaveManagement.select_save()
        if new_save_file is None:
            return stop
        save_file.load_save_file(new_save_file)
        core.core_data.init_data()
        color.ColoredText.localize("load_save_success")
        return False

    @staticmethod
    def convert_save_cc(save_file: core.SaveFile):
        color.ColoredText.localize("cc_warning", current=save_file.cc)
        ccs_to_select = core.CountryCode.get_all()
        cc = core.CountryCode.select_from_ccs(ccs_to_select)
        if cc is None:
            return
        save_file.set_cc(cc)
        core.ServerHandler(save_file).create_new_account()
        core.core_data.init_data()
        color.ColoredText.localize("country_code_set", cc=cc)

    @staticmethod
    def convert_save_gv(save_file: core.SaveFile):
        color.ColoredText.localize(
            "gv_warning", current=save_file.game_version.to_string()
        )
        try:
            gv = core.GameVersion.from_string(
                color.ColoredInput().localize("game_version_dialog").strip()
            )
        except ValueError:
            color.ColoredText.localize("invalid_game_version")
            return
        save_file.set_gv(gv)
        core.core_data.init_data()
        color.ColoredText.localize("game_version_set", version=gv.to_string())



from __future__ import annotations
import enum
from typing import Any
from bcsfe import core
from bcsfe.cli import color, dialog_creator
import requests


class ConfigKey(enum.Enum):
    UPDATE_TO_BETA = "update_to_beta"
    SHOW_UPDATE_MESSAGE = "show_update_message"
    LOCALE = "locale"
    SHOW_MISSING_LOCALE_KEYS = "show_missing_locale_keys"
    DISABLE_MAXES = "disable_maxes"
    MAX_BACKUPS = "max_backups"
    THEME = "theme"
    RESET_CAT_DATA = "reset_cat_data"
    SET_CAT_CURRENT_FORMS = "set_cat_current_forms"
    STRICT_UPGRADE = "strict_upgrade"
    SEPARATE_CAT_EDIT_OPTIONS = "separate_cat_edit_options"
    STRICT_BAN_PREVENTION = "strict_ban_prevention"
    MAX_REQUEST_TIMEOUT = "max_request_timeout"
    GAME_DATA_REPO = "game_data_repo"
    FORCE_LANG_GAME_DATA = "force_lang_game_data"
    CLEAR_TUTORIAL_ON_LOAD = "clear_tutorial_on_load"
    REMOVE_BAN_MESSAGE_ON_LOAD = "remove_ban_message_on_load"
    UNLOCK_CAT_ON_EDIT = "unlock_cat_on_edit"
    USE_FILE_DIALOG = "use_file_dialog"
    ADB_PATH = "adb_path"
    IGNORE_PARSE_ERROR = "ignore_parse_error"
    USE_WAYDROID = "use_waydroid"
    USE_PKEXEC_WAYDROID = "use_pkexec_waydroid"


class Config:
    def __init__(self, path: core.Path | None, print_yaml_err: bool = True):
        if path is None:
            path = Config.get_config_path()
        config = core.YamlFile(path, print_yaml_err)
        self.config: dict[ConfigKey, Any] = {}
        for key, value in config.yaml.items():
            try:
                self.config[ConfigKey(key)] = value
            except ValueError:
                pass
        self.config_object = config
        self.initialize_config()

    @staticmethod
    def get_config_path() -> core.Path:
        return core.Path.get_documents_folder().add("config.yaml")

    def __getitem__(self, key: ConfigKey) -> Any:
        return self.config[key]

    def __setitem__(self, key: ConfigKey, value: Any) -> None:
        self.config[key] = value

    def __contains__(self, key: ConfigKey) -> bool:
        return key in self.config

    @staticmethod
    def get_defaults() -> dict[ConfigKey, Any]:
        initial_values = {
            ConfigKey.UPDATE_TO_BETA: False,
            ConfigKey.SHOW_UPDATE_MESSAGE: True,
            ConfigKey.LOCALE: "en",
            ConfigKey.SHOW_MISSING_LOCALE_KEYS: False,
            ConfigKey.DISABLE_MAXES: False,
            ConfigKey.MAX_BACKUPS: 50,
            ConfigKey.THEME: "default",
            ConfigKey.RESET_CAT_DATA: True,
            ConfigKey.SET_CAT_CURRENT_FORMS: True,
            ConfigKey.STRICT_UPGRADE: False,
            ConfigKey.SEPARATE_CAT_EDIT_OPTIONS: True,
            ConfigKey.STRICT_BAN_PREVENTION: False,
            ConfigKey.MAX_REQUEST_TIMEOUT: 30,
            ConfigKey.GAME_DATA_REPO: "https://git.battlecatsmodding.org/fieryhenry/BCData/raw/branch/main/metadata.json",
            ConfigKey.FORCE_LANG_GAME_DATA: False,
            ConfigKey.CLEAR_TUTORIAL_ON_LOAD: True,
            ConfigKey.REMOVE_BAN_MESSAGE_ON_LOAD: True,
            ConfigKey.UNLOCK_CAT_ON_EDIT: True,
            ConfigKey.USE_FILE_DIALOG: True,
            ConfigKey.ADB_PATH: "adb",
            ConfigKey.IGNORE_PARSE_ERROR: False,
            ConfigKey.USE_WAYDROID: False,
            ConfigKey.USE_PKEXEC_WAYDROID: True,
        }
        return initial_values

    def get_default(self, key: ConfigKey) -> Any:
        value = Config.get_defaults()[key]
        return value

    def set_default(self, key: ConfigKey):
        value = self.get_default(key)
        self.config[key] = value
        self.save()
        return value

    def initialize_config(self):
        initial_values = Config.get_defaults()

        for key, value in initial_values.items():
            if key not in self.config:
                self.config[key] = value
        self.save()

    def save(self):
        for key, value in self.config.items():
            self.config_object.yaml[key.value] = value
        self.config_object.save()

    def get(self, key: ConfigKey) -> Any:
        value = self.config[key]
        if value is None:
            return self.set_default(key)
        return value

    def get_game_data_repo(self, fix_old_repo: bool = True) -> str:
        if fix_old_repo and self.get_str(ConfigKey.GAME_DATA_REPO) in [
            "https://raw.githubusercontent.com/fieryhenry/BCData/master/",
            "https://git.fyhenry.uk/henry/BCData/raw/branch/main/info.json",
        ]:
            self.set(
                ConfigKey.GAME_DATA_REPO, self.get_default(ConfigKey.GAME_DATA_REPO)
            )
        return self.get_str(ConfigKey.GAME_DATA_REPO)

    def get_str(self, key: ConfigKey) -> str:
        value = self.get(key)
        if not isinstance(value, str):
            return self.set_default(key)
        return value

    def get_bool(self, key: ConfigKey) -> bool:
        value = self.get(key)
        if not isinstance(value, bool):
            return self.set_default(key)
        return value

    def get_int(self, key: ConfigKey) -> int:
        value = self.get(key)
        if not isinstance(value, int):
            return self.set_default(key)
        return value

    def reset(self):
        self.config.clear()
        self.config_object.remove()
        self.initialize_config()

    def set(self, key: ConfigKey, value: Any):
        self.config[key] = value
        self.save()

    def get_bool_text(self, value: bool) -> str:
        if value:
            return core.core_data.local_manager.get_key("enabled")
        return core.core_data.local_manager.get_key("disabled")

    def get_full_input_localized(
        self, key: ConfigKey, current_value: str, default_value: str
    ) -> str:
        return core.core_data.local_manager.get_key(
            "config_full",
            key_desc=core.core_data.local_manager.get_key(
                Config.get_desc_key(key),
                current_value=current_value,
                default_value=default_value,
                escape=False,
            ),
            escape=False,
        )

    def edit_bool(self, key: ConfigKey):
        value = self.get_bool(key)
        color.ColoredText(
            self.get_full_input_localized(
                key,
                self.get_bool_text(value),
                self.get_bool_text(self.get_default(key)),
            ),
        )
        choice = dialog_creator.ChoiceInput(
            ["enable", "disable"],
            ["enable", "disable"],
            [],
            {},
            "enable_disable_dialog",
            True,
        ).single_choice()
        if choice is None:
            return
        choice -= 1
        if choice == 0:
            value = True
        elif choice == 1:
            value = False
        self.set(key, value)
        print()
        color.ColoredText.localize(
            "config_success",
        )

    @staticmethod
    def get_desc_key(key: ConfigKey) -> str:
        return key.value + "_desc"

    def edit_int(self, key: ConfigKey):
        text = self.get_full_input_localized(
            key, str(self.get_int(key)), str(self.get_default(key))
        )
        color.ColoredText.localize(text)
        value = dialog_creator.SingleEditor(
            key.value, self.get_int(key), signed=False, localized_item=True
        ).edit()
        self.set(key, value)

        color.ColoredText.localize(
            "config_success",
        )

    def edit_game_data_repo(self):
        text = self.get_full_input_localized(
            ConfigKey.GAME_DATA_REPO,
            self.get_str(ConfigKey.GAME_DATA_REPO),
            self.get_default(ConfigKey.GAME_DATA_REPO),
        )
        color.ColoredText.localize(text)
        value = dialog_creator.StringInput().get_input_locale(
            "game_data_repo_dialog", {}
        )
        if value is None:
            value = self.get_default(ConfigKey.GAME_DATA_REPO)
        color.ColoredText.localize("validating_game_repo")
        try:
            resp = core.RequestHandler(value).get()
        except (requests.exceptions.MissingSchema, requests.exceptions.InvalidSchema):
            color.ColoredText.localize("invalid_url")
            return
        if resp is None:
            color.ColoredText.localize("no_internet_or_connection_error")
            return
        if resp.status_code != 200:
            color.ColoredText.localize(
                "invalid_response", response_code=resp.status_code
            )
            return
        self.set(ConfigKey.GAME_DATA_REPO, value)
        color.ColoredText.localize(
            "config_success",
        )

    def edit_str(self, key: ConfigKey):
        text = self.get_full_input_localized(
            key,
            self.get_str(key),
            self.get_default(key),
        )

        color.ColoredText.localize(text)

        str_val = core.core_data.local_manager.get_key(key.value)

        value = dialog_creator.StringInput().get_input_locale(
            "string_config_dialog", {"val": str_val}
        )
        if value is None:
            return

        self.set(key, value)
        color.ColoredText.localize("config_success")

    def edit_locale(self):
        text = self.get_full_input_localized(
            ConfigKey.LOCALE,
            self.get_str(ConfigKey.LOCALE),
            self.get_default(ConfigKey.LOCALE),
        )
        color.ColoredText.localize(text)
        all_locales = core.LocalManager.get_all_locales()
        options = all_locales.copy() + ["add_locale", "remove_locale"]
        value = dialog_creator.ChoiceInput.from_reduced(
            options,
            dialog="locale_dialog",
            single_choice=True,
        ).single_choice()
        if value is None:
            return
        value -= 1
        if value == len(all_locales) + 1:  # remove_locale
            options: list[str] = []
            for locale in all_locales:
                if locale.startswith("ext-"):
                    options.append(locale)

            if not options:
                color.ColoredText.localize(
                    "no_external_locales",
                )
                return

            options.append("cancel")

            choices, _ = dialog_creator.ChoiceInput.from_reduced(
                options, dialog="locale_remove_dialog"
            ).multiple_choice()
            if choices is None:
                return
            for choice in choices:
                if choice == len(options) - 1:
                    return
                core.LocalManager.remove_locale(options[choice])
                color.ColoredText.localize(
                    "locale_removed",
                    locale_name=options[choice],
                )
            return
        elif value == len(all_locales):  # add_locale
            if not core.GitHandler.is_git_installed():
                color.ColoredText.localize(
                    "git_not_installed",
                )
                return
            git_repo = color.ColoredInput().localize("enter_locale_git_repo").strip()
            external_locale = core.ExternalLocale.from_git_repo(git_repo)
            if external_locale is None:
                color.ColoredText.localize(
                    "invalid_git_repo",
                )
                return
            locale_name = external_locale.get_full_name()
            if locale_name in all_locales:
                if not dialog_creator.YesNoInput().get_input_once(
                    "locale_already_exists",
                    {"locale_name": locale_name},
                ):
                    color.ColoredText.localize(
                        "locale_cancelled",
                    )
                    return

            external_locale.save()

            value = locale_name
            color.ColoredText.localize(
                "locale_added",
            )
        else:
            value = all_locales[value]
        self.set(ConfigKey.LOCALE, value)
        color.ColoredText.localize(
            "locale_changed",
            locale_name=value,
        )
        color.ColoredText.localize(
            "config_success",
        )

    def edit_theme(self):
        themes = core.ThemeHandler.get_all_themes()
        current_theme = self.get_str(ConfigKey.THEME)
        if current_theme not in themes:
            current_theme = "default"
        text = self.get_full_input_localized(
            ConfigKey.THEME,
            current_theme,
            self.get_default(ConfigKey.THEME),
        )
        color.ColoredText.localize(text)
        options = themes.copy() + ["add_theme", "remove_theme"]
        value = dialog_creator.ChoiceInput.from_reduced(
            options,
            dialog="theme_dialog",
            single_choice=True,
        ).single_choice()
        if value is None:
            return
        value -= 1
        if value == len(themes) + 1:  # remove_theme
            options: list[str] = []
            for theme in themes:
                if theme.startswith("ext-"):
                    options.append(theme)

            if not options:
                color.ColoredText.localize(
                    "no_external_themes",
                )
                return

            options.append("cancel")

            choices, _ = dialog_creator.ChoiceInput.from_reduced(
                options, dialog="theme_remove_dialog"
            ).multiple_choice()
            if choices is None:
                return
            for choice in choices:
                if choice == len(options) - 1:
                    return
                core.ThemeHandler.remove_theme(options[choice])
                color.ColoredText.localize(
                    "theme_removed",
                    theme_name=options[choice],
                )
            return
        elif value == len(themes):  # add_theme
            if not core.GitHandler.is_git_installed():
                color.ColoredText.localize(
                    "git_not_installed",
                )
                return
            git_repo = color.ColoredInput().localize("enter_theme_git_repo").strip()
            external_theme = core.ExternalTheme.from_git_repo(git_repo)
            if external_theme is None:
                color.ColoredText.localize(
                    "invalid_git_repo",
                )
                return
            theme_name = external_theme.get_full_name()
            if theme_name in themes:
                if not dialog_creator.YesNoInput().get_input_once(
                    "theme_already_exists",
                    {"theme_name": theme_name},
                ):
                    color.ColoredText.localize(
                        "theme_cancelled",
                    )
                    return

            external_theme.save()

            value = theme_name
            color.ColoredText.localize(
                "theme_added",
            )
        else:
            value = themes[value]
        self.set(ConfigKey.THEME, value)
        color.ColoredText.localize(
            "theme_changed",
            theme_name=value,
        )

    @staticmethod
    def edit_config(_: Any = None):
        config = core.core_data.config
        features = list(ConfigKey)

        choice = dialog_creator.ChoiceInput.from_reduced(
            [key.value for key in features],
            dialog="config_dialog",
            single_choice=True,
        ).single_choice()
        if choice is None:
            return
        choice -= 1
        feature = features[choice]
        print()
        if isinstance(config.get(feature), bool):
            core.core_data.config.edit_bool(feature)
        elif isinstance(config.get(feature), int):
            core.core_data.config.edit_int(feature)
        elif feature == ConfigKey.LOCALE:
            core.core_data.config.edit_locale()
        elif feature == ConfigKey.THEME:
            core.core_data.config.edit_theme()
        elif feature == ConfigKey.GAME_DATA_REPO:
            core.core_data.config.edit_game_data_repo()
        elif isinstance(config.get(feature), str):
            core.core_data.config.edit_str(feature)
        print()



from __future__ import annotations
from typing import Any
from bcsfe import core


class ClientInfo:
    def __init__(self, cc: core.CountryCode, gv: core.GameVersion):
        self.cc = cc
        self.gv = gv

    @staticmethod
    def from_save_file(save_file: core.SaveFile):
        return ClientInfo(save_file.cc, save_file.game_version)

    def get_client_info(self) -> dict[str, Any]:
        cc = self.cc.get_client_info_code()

        data = {
            "clientInfo": {
                "client": {
                    "countryCode": cc,
                    "version": self.gv.game_version,
                },
                "device": {
                    "model": "SM-G955F",
                },
                "os": {
                    "type": "android",
                    "version": "9",
                },
            },
            "nonce": core.Random.get_hex_string(32),
        }
        return data



from __future__ import annotations
from typing import Any
from bcsfe import core


class SkillLevel:
    def __init__(
        self,
        id: int,
        levels: list[int],
    ):
        self.id = id
        self.levels = levels

    def get_total_levels(self) -> int:
        return len(self.levels)

    @staticmethod
    def from_row(row: core.Row):
        id = row[0].to_int()
        levels = row[1:].to_int_list()
        return SkillLevel(id, levels)


class SkillLevelData:
    def __init__(self, levels: list[SkillLevel] | None):
        self.levels = levels

    @staticmethod
    def from_game_data(save_file: core.SaveFile) -> SkillLevelData | None:
        gdg = core.core_data.get_game_data_getter(save_file)
        data = gdg.download("DataLocal", "SkillLevel.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        levels: list[SkillLevel] = []
        for line in csv.lines[1:]:
            levels.append(SkillLevel.from_row(line))
        return SkillLevelData(levels)

    def get_skill_level(self, id: int) -> SkillLevel | None:
        if self.levels is None:
            return None
        for level in self.levels:
            if level.id == id:
                return level
        return None


class Skill:
    def __init__(
        self,
        ability_id: int,
        max_lv: int,
        min1: int,
        max1: int,
        min2: int,
        max2: int,
        min3: int,
        max3: int,
        min4: int,
        max4: int,
        text_id: int,
        lvid: int,
        name_id: int,
        limit: int,
    ):
        self.ability_id = ability_id
        self.max_lv = max_lv
        self.min1 = min1
        self.max1 = max1
        self.min2 = min2
        self.max2 = max2
        self.min3 = min3
        self.max3 = max3
        self.min4 = min4
        self.max4 = max4
        self.text_id = text_id
        self.lvid = lvid
        self.name_id = name_id
        self.limit = limit


class CatSkill:
    def __init__(
        self,
        cat_id: int,
        type_id: int,
        skills: list[Skill],
    ):
        self.cat_id = cat_id
        self.type_id = type_id
        self.skills = skills

    @staticmethod
    def from_row(row: core.Row):
        cat_id = row[0].to_int()
        type_id = row[1].to_int()
        skills: list[Skill] = []
        for i in range(2, len(row), 14):
            skill = Skill(
                row[i].to_int(),
                row[i + 1].to_int(),
                row[i + 2].to_int(),
                row[i + 3].to_int(),
                row[i + 4].to_int(),
                row[i + 5].to_int(),
                row[i + 6].to_int(),
                row[i + 7].to_int(),
                row[i + 8].to_int(),
                row[i + 9].to_int(),
                row[i + 10].to_int(),
                row[i + 11].to_int(),
                row[i + 12].to_int(),
                row[i + 13].to_int(),
            )
            skills.append(skill)
        return CatSkill(cat_id, type_id, skills)


class CatSkills:
    def __init__(self, skills: dict[int, CatSkill]):
        self.skills = skills

    @staticmethod
    def from_game_data(save_file: core.SaveFile) -> CatSkills | None:
        gdg = core.core_data.get_game_data_getter(save_file)
        data = gdg.download("DataLocal", "SkillAcquisition.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        skills: dict[int, CatSkill] = {}
        for line in csv.lines[1:]:
            skill = CatSkill.from_row(line)
            skills[skill.cat_id] = skill
        return CatSkills(skills)

    def get_cat_skill(self, cat_id: int) -> CatSkill | None:
        return self.skills.get(cat_id)


class SkillNames:
    def __init__(self, names: dict[int, str]):
        self.names = names

    @staticmethod
    def from_game_data(save_file: core.SaveFile) -> SkillNames | None:
        gdg = core.core_data.get_game_data_getter(save_file)
        data = gdg.download("resLocal", "SkillDescriptions.csv")
        if data is None:
            return None
        csv = core.CSV(
            data, delimiter=core.Delimeter.from_country_code_res(save_file.cc)
        )
        names: dict[int, str] = {}
        for line in csv.lines[1:]:
            names[line[0].to_int()] = line[1].to_str()
        return SkillNames(names)

    def get_skill_name(self, skill_id: int) -> str | None:
        return self.names.get(skill_id)


class TalentData:
    def __init__(
        self,
        skill_names: SkillNames,
        skill_levels: SkillLevelData,
        cats: CatSkills,
    ):
        self.skill_names = skill_names
        self.skill_levels = skill_levels
        self.cats = cats

    @staticmethod
    def from_game_data(save_file: core.SaveFile) -> TalentData | None:
        skill_names = SkillNames.from_game_data(save_file)
        skill_levels = SkillLevelData.from_game_data(save_file)
        cats = CatSkills.from_game_data(save_file)
        if skill_names is None or skill_levels is None or cats is None:
            return None

        return TalentData(skill_names, skill_levels, cats)

    def get_skill_name(self, skill_id: int) -> str | None:
        return self.skill_names.get_skill_name(skill_id)

    def get_skill_level(self, skill_id: int) -> SkillLevel | None:
        return self.skill_levels.get_skill_level(skill_id)

    def get_cat_skill(self, cat_id: int) -> CatSkill | None:
        return self.cats.get_cat_skill(cat_id)

    def get_skill_from_cat(self, cat_id: int, skill_id: int) -> Skill | None:
        cat_skill = self.get_cat_skill(cat_id)
        if cat_skill is None:
            return None
        for skill in cat_skill.skills:
            if skill.ability_id == skill_id:
                return skill
        return None

    def get_talent_from_cat_skill(self, cat: core.Cat, skill_id: int) -> Talent | None:
        talents = cat.talents
        if talents is None:
            return None
        for talent in talents:
            if talent.id == skill_id:
                return talent
        return None

    def get_cat_skill_name(self, cat_id: int, skill_id: int) -> str | None:
        skill = self.get_skill_from_cat(cat_id, skill_id)
        if skill is None:
            return None
        return self.get_skill_name(skill.text_id)

    def get_cat_skill_level(self, cat_id: int, skill_id: int) -> SkillLevel | None:
        skill = self.get_skill_from_cat(cat_id, skill_id)
        if skill is None:
            return None
        return self.get_skill_level(skill.lvid)

    def get_cat_talents(
        self, cat: core.Cat
    ) -> tuple[list[str], list[int], list[int], list[int]] | None:
        talent_data_cat = self.get_cat_skill(cat.id)
        if talent_data_cat is None or cat.talents is None:
            return None
        # save_talent_data = cat.talents
        talent_names: list[str] = []
        max_levels: list[int] = []
        current_levels: list[int] = []
        ids: list[int] = []
        for skill in talent_data_cat.skills:
            name = self.get_skill_name(skill.text_id)
            talent = self.get_talent_from_cat_skill(cat, skill.ability_id)
            if name is None or talent is None:
                continue

            max_level = skill.max_lv
            if max_level == 0:
                max_level = 1

            max_levels.append(max_level)
            talent_names.append(name.split("<br>")[0])
            current_levels.append(talent.level)
            ids.append(skill.ability_id)

        return talent_names, max_levels, current_levels, ids


class Talent:
    def __init__(self, id: int, level: int):
        self.id = id
        self.level = level

    @staticmethod
    def init() -> Talent:
        return Talent(0, 0)

    def reset(self):
        self.level = 0

    @staticmethod
    def read(stream: core.Data):
        return Talent(stream.read_int(), stream.read_int())

    def write(self, stream: core.Data):
        stream.write_int(self.id)
        stream.write_int(self.level)

    def serialize(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "level": self.level,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Talent:
        return Talent(
            data["id"],
            data["level"],
        )

    def __repr__(self):
        return f"Talent({self.id}, {self.level})"

    def __str__(self):
        return self.__repr__()


class NyankoPictureBookCatData:
    def __init__(
        self,
        cat_id: int,
        is_displayed_in_catguide: bool,
        limited: bool,
        total_forms: int,
        hint_display_type: int,
        scale_0: int,
        scale_1: int,
        scale_2: int,
        scale_3: int,
    ):
        self.cat_id = cat_id
        self.is_displayed_in_catguide = is_displayed_in_catguide
        self.limited = limited
        self.total_forms = total_forms
        self.hint_display_type = hint_display_type
        self.scale_0 = scale_0
        self.scale_1 = scale_1
        self.scale_2 = scale_2
        self.scale_3 = scale_3


class NyankoPictureBook:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.cats = self.get_cats()

    def get_cats(self) -> list[NyankoPictureBookCatData] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "nyankoPictureBookData.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        cats: list[NyankoPictureBookCatData] = []
        for i, line in enumerate(csv):
            cat = NyankoPictureBookCatData(
                i,
                line[0].to_bool(),
                line[1].to_bool(),
                line[2].to_int(),
                line[3].to_int(),
                line[4].to_int(),
                line[5].to_int(),
                line[6].to_int(),
                line[7].to_int(),
            )
            cats.append(cat)
        return cats

    def get_cat(self, cat_id: int) -> NyankoPictureBookCatData | None:
        if self.cats is None:
            return None
        for cat in self.cats:
            if cat.cat_id == cat_id:
                return cat
        return None

    def get_obtainable_cats(self) -> list[NyankoPictureBookCatData] | None:
        if self.cats is None:
            return None
        return [cat for cat in self.cats if cat.is_displayed_in_catguide]


class EvolveItem:
    """Represents an item used to evolve a unit."""

    def __init__(
        self,
        item_id: int,
        amount: int,
    ):
        """Initializes a new EvolveItem object.

        Args:
            item_id (int): The ID of the item.
            amount (int): The amount of the item.
        """
        self.item_id = item_id
        self.amount = amount

    def __str__(self) -> str:
        """Gets a string representation of the EvolveItem object.

        Returns:
            str: The string representation of the EvolveItem object.
        """
        return f"{self.item_id}:{self.amount}"

    def __repr__(self) -> str:
        """Gets a string representation of the EvolveItem object.

        Returns:
            str: The string representation of the EvolveItem object.
        """
        return str(self)


class EvolveItems:
    """Represents the items used to evolve a unit."""

    def __init__(self, evolve_items: list[EvolveItem]):
        """Initializes a new EvolveItems object.

        Args:
            evolve_items (list[EvolveItem]): The items used to evolve a unit.
        """
        self.evolve_items = evolve_items

    @staticmethod
    def from_unit_buy_list(raw_data: core.Row, start_index: int) -> EvolveItems:
        """Creates a new EvolveItems object from a row from unitbuy.csv.

        Args:
            raw_data (core.Row): The row from unitbuy.csv.

        Returns:
            EvolveItems: The EvolveItems object.
        """
        items: list[EvolveItem] = []
        for i in range(5):
            item_id = raw_data[start_index + i * 2].to_int()
            amount = raw_data[start_index + 1 + i * 2].to_int()
            items.append(EvolveItem(item_id, amount))
        return EvolveItems(items)


class UnitBuyCatData:
    def __init__(self, id: int, raw_data: core.Row):
        self.id = id
        self.assign(raw_data)

    def assign(self, raw_data: core.Row):
        self.stage_unlock = raw_data[0].to_int()
        self.purchase_cost = raw_data[1].to_int()
        self.upgrade_costs = [cost.to_int() for cost in raw_data[2:12]]
        self.unlock_source = raw_data[12].to_int()
        self.rarity = raw_data[13].to_int()
        self.position_order = raw_data[14].to_int()
        self.chapter_unlock = raw_data[15].to_int()
        self.sell_price = raw_data[16].to_int()
        self.gatya_rarity = raw_data[17].to_int()
        self.original_max_levels = raw_data[18].to_int(), raw_data[19].to_int()
        self.force_true_form_level = raw_data[20].to_int()
        self.second_form_unlock_level = raw_data[21].to_int()
        self.unknown_22 = raw_data[22].to_int()
        self.tf_id = raw_data[23].to_int()
        self.ff_id = raw_data[24].to_int()
        self.evolve_level_tf = raw_data[25].to_int()
        self.evolve_level_ff = raw_data[26].to_int()
        self.evolve_cost_tf = raw_data[27].to_int()
        self.evolve_items_tf = EvolveItems.from_unit_buy_list(raw_data, 28)
        self.evolve_cost_ff = raw_data[38].to_int()
        self.evolve_items_ff = EvolveItems.from_unit_buy_list(raw_data, 39)
        self.max_upgrade_level_no_catseye = raw_data[49].to_int()
        self.max_upgrade_level_catseye = raw_data[50].to_int()
        self.max_plus_upgrade_level = raw_data[51].to_int()
        self.unknown_52 = raw_data[52].to_int()
        self.unknown_53 = raw_data[53].to_int()
        self.unknown_54 = raw_data[54].to_int()
        self.unknown_55 = raw_data[55].to_int()
        self.catseye_usage_pattern = raw_data[56].to_int()
        self.game_version = raw_data[57].to_int()
        self.np_sell_price = raw_data[58].to_int()
        self.unknwon_59 = raw_data[59].to_int()
        self.unknown_60 = raw_data[60].to_int()
        self.egg_value = raw_data[61].to_int()
        self.egg_id = raw_data[62].to_int()


class UnitBuy:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.unit_buy = self.read_unit_buy()

    def read_unit_buy(self) -> list[UnitBuyCatData] | None:
        unit_buy: list[UnitBuyCatData] = []
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "unitbuy.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        for i, line in enumerate(csv):
            unit_buy.append(UnitBuyCatData(i, line))
        return unit_buy

    def get_unit_buy(self, id: int) -> UnitBuyCatData | None:
        if self.unit_buy is None:
            return None
        try:
            return self.unit_buy[id]
        except IndexError:
            return None

    def get_cat_rarity(self, id: int) -> int:
        unit_buy = self.get_unit_buy(id)
        if unit_buy is None:
            return -1
        return unit_buy.rarity


class UnitLimitCatData:
    def __init__(self, cat_id: int, values: list[int]):
        self.cat_id = cat_id
        self.values = values


class UnitLimit:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.unit_limit = self.read_unit_limit()

    def read_unit_limit(self) -> list[UnitLimitCatData] | None:
        unit_limit: list[UnitLimitCatData] = []
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "unitlimit.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        for i, line in enumerate(csv):
            unit_limit.append(UnitLimitCatData(i, line.to_int_list()))
        return unit_limit

    def get_unit_limit(self, id: int) -> UnitLimitCatData | None:
        if self.unit_limit is None:
            return None

        try:
            return self.unit_limit[id]
        except IndexError:
            return None


class Cat:
    def __init__(self, id: int, unlocked: int):
        self.id = id
        self.unlocked = unlocked
        self.talents: list[Talent] | None = None
        self.upgrade: core.Upgrade = core.Upgrade.init()
        self.current_form: int = 0
        self.unlocked_forms: int = 0
        self.gatya_seen: int = 0
        self.max_upgrade_level: core.Upgrade = core.Upgrade.init()
        self.catguide_collected: bool = False
        self.fourth_form: int = 0
        self.catseyes_used: int = 0

        self.names: list[str] | None = None

    def get_talent_from_id(self, id: int) -> Talent | None:
        for talent in self.talents or []:
            if talent.id == id:
                return talent
        return None

    def unlock(self, save_file: core.SaveFile):
        self.unlocked = 1
        self.gatya_seen = 1
        core.core_data.get_chara_drop(save_file).unlock_drops_from_cat_id(self.id)
        save_file.unlock_equip_menu()

    def remove(self, reset: bool = False, save_file: core.SaveFile | None = None):
        self.unlocked = 0
        if reset:
            self.reset()
            if save_file is not None:
                save_file.cats.chara_new_flags[self.id] = 0
                core.core_data.get_chara_drop(save_file).remove_drops_from_cat_id(
                    self.id
                )

    def true_form(self, save_file: core.SaveFile, set_current_form: bool = True):
        self.set_form(2, save_file, set_current_form)

    def set_form(
        self, form: int, save_file: core.SaveFile, set_current_form: bool = True
    ):
        if core.core_data.config.get_bool(core.ConfigKey.UNLOCK_CAT_ON_EDIT):
            self.unlock(save_file)
        self.unlocked_forms = form + 1
        if set_current_form:
            self.current_form = form

    def set_form_true(
        self,
        save_file: core.SaveFile,
        total_forms: int,
        set_current_form: bool = True,
        fourth_form: bool = False,
    ):
        if total_forms == 4 and self.unlocked_forms == 3 and fourth_form:
            self.unlock_fourth_form(save_file, set_current_form)
        elif total_forms >= 3:
            self.true_form(save_file, set_current_form)
        elif total_forms == 2:
            self.unlocked_forms = 0
            self.current_form = 1
        else:
            self.unlocked_forms = 0
            self.current_form = 0

    def remove_true_form(self):
        self.unlocked_forms = 0
        self.current_form = min(self.current_form, 1)
        self.fourth_form = 0

    def unlock_fourth_form(
        self, save_file: core.SaveFile, set_current_form: bool = True
    ):
        if set_current_form:
            self.current_form = 3
        if core.core_data.config.get_bool(core.ConfigKey.UNLOCK_CAT_ON_EDIT):
            self.unlock(save_file)
        self.fourth_form = 2

    def remove_fourth_form(self):
        self.current_form = min(self.current_form, 2)
        self.fourth_form = 0

    def set_upgrade(
        self,
        save_file: core.SaveFile,
        upgrade: core.Upgrade,
        only_plus: bool = False,
    ):
        if core.core_data.config.get_bool(core.ConfigKey.UNLOCK_CAT_ON_EDIT):
            self.unlock(save_file)
        base = upgrade.base
        plus = upgrade.plus
        if base != -1 and not only_plus:
            self.upgrade.base = upgrade.get_random_base()
        if plus != -1:
            self.upgrade.plus = upgrade.get_random_plus()

    def upgrade_base(self, save_file: core.SaveFile):
        if core.core_data.config.get_bool(core.ConfigKey.UNLOCK_CAT_ON_EDIT):
            self.unlock(save_file)
        self.upgrade.upgrade()

    def reset(self):
        self.unlocked = 0
        self.current_form = 0
        self.unlocked_forms = 0
        self.gatya_seen = 0
        self.catguide_collected = False
        self.fourth_form = 0
        self.catseyes_used = 0
        self.upgrade.reset()
        for talent in self.talents or []:
            talent.reset()

    @staticmethod
    def init(id: int) -> Cat:
        return Cat(id, 0)

    @staticmethod
    def read_unlocked(id: int, stream: core.Data):
        return Cat(id, stream.read_int())

    def write_unlocked(self, stream: core.Data):
        stream.write_int(self.unlocked)

    def read_upgrade(self, stream: core.Data):
        self.upgrade = core.Upgrade.read(stream)

    def write_upgrade(self, stream: core.Data):
        self.upgrade.write(stream)

    def read_current_form(self, stream: core.Data):
        self.current_form = stream.read_int()

    def write_current_form(self, stream: core.Data):
        stream.write_int(self.current_form)

    def read_unlocked_forms(self, stream: core.Data):
        self.unlocked_forms = stream.read_int()

    def write_unlocked_forms(self, stream: core.Data):
        stream.write_int(self.unlocked_forms)

    def read_gatya_seen(self, stream: core.Data):
        self.gatya_seen = stream.read_int()

    def write_gatya_seen(self, stream: core.Data):
        stream.write_int(self.gatya_seen)

    def read_max_upgrade_level(self, stream: core.Data):
        level = core.Upgrade.read(stream)
        self.max_upgrade_level = level

    def write_max_upgrade_level(self, stream: core.Data):
        self.max_upgrade_level.write(stream)

    def read_catguide_collected(self, stream: core.Data):
        self.catguide_collected = stream.read_bool()

    def write_catguide_collected(self, stream: core.Data):
        stream.write_bool(self.catguide_collected)

    def read_fourth_form(self, stream: core.Data):
        self.fourth_form = stream.read_int()

    def write_fourth_form(self, stream: core.Data):
        stream.write_int(self.fourth_form)

    def read_catseyes_used(self, stream: core.Data):
        self.catseyes_used = stream.read_int()

    def write_catseyes_used(self, stream: core.Data):
        stream.write_int(self.catseyes_used)

    def serialize(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "unlocked": self.unlocked,
            "upgrade": self.upgrade.serialize(),
            "current_form": self.current_form,
            "unlocked_forms": self.unlocked_forms,
            "gatya_seen": self.gatya_seen,
            "max_upgrade_level": self.max_upgrade_level.serialize(),
            "catguide_collected": self.catguide_collected,
            "fourth_form": self.fourth_form,
            "catseyes_used": self.catseyes_used,
            "talents": (
                [talent.serialize() for talent in self.talents]
                if self.talents is not None
                else None
            ),
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Cat:
        cat = Cat(data["id"], data["unlocked"])
        cat.upgrade = core.Upgrade.deserialize(data["upgrade"])
        cat.current_form = data["current_form"]
        cat.unlocked_forms = data["unlocked_forms"]
        cat.gatya_seen = data["gatya_seen"]
        cat.max_upgrade_level = core.Upgrade.deserialize(data["max_upgrade_level"])
        cat.catguide_collected = data["catguide_collected"]
        cat.fourth_form = data["fourth_form"]
        cat.catseyes_used = data["catseyes_used"]
        cat.talents = (
            [Talent.deserialize(talent) for talent in data["talents"]]
            if data["talents"] is not None
            else None
        )
        return cat

    def __repr__(self) -> str:
        return f"Cat(id={self.id}, unlocked={self.unlocked}, upgrade={self.upgrade}, current_form={self.current_form}, unlocked_forms={self.unlocked_forms}, gatya_seen={self.gatya_seen}, max_upgrade_level={self.max_upgrade_level}, catguide_collected={self.catguide_collected}, fourth_form={self.fourth_form}, catseyes_used={self.catseyes_used}, talents={self.talents})"

    def __str__(self) -> str:
        return self.__repr__()

    def read_talents(self, stream: core.Data):
        self.talents = []
        for _ in range(stream.read_int()):
            self.talents.append(Talent.read(stream))

    def write_talents(self, stream: core.Data):
        if self.talents is None:
            return
        stream.write_int(len(self.talents))
        for talent in self.talents:
            talent.write(stream)

    def get_names_cls(self, save_file: core.SaveFile) -> list[str] | None:
        if self.names is None:
            self.names = Cat.get_names(self.id, save_file)
        return self.names

    @staticmethod
    def get_names(
        id: int,
        save_file: core.SaveFile,
    ) -> list[str] | None:
        file_name = f"Unit_Explanation{id + 1}_{core.core_data.get_lang(save_file)}.csv"
        data = core.core_data.get_game_data_getter(save_file).download(
            "resLocal", file_name
        )
        if data is None:
            return None
        csv = core.CSV(
            data,
            core.Delimeter.from_country_code_res(save_file.cc),
            remove_empty=False,
        )
        names: list[str] = []
        for line in csv.lines:
            names.append(line[0].to_str())

        return names


class StorageItem:
    def __init__(self, item_id: int):
        self.item_id = item_id
        self.item_type = 0

    @staticmethod
    def from_cat(cat_id: int) -> StorageItem:
        item = StorageItem(cat_id)
        item.item_type = 1
        return item

    @staticmethod
    def from_special_skill(special_skill_id: int) -> StorageItem:
        item = StorageItem(special_skill_id)
        item.item_type = 2
        return item

    @staticmethod
    def init() -> StorageItem:
        return StorageItem(0)

    @staticmethod
    def read_item_id(stream: core.Data):
        return StorageItem(stream.read_int())

    def write_item_id(self, stream: core.Data):
        stream.write_int(self.item_id)

    def read_item_type(self, stream: core.Data):
        self.item_type = stream.read_int()

    def write_item_type(self, stream: core.Data):
        stream.write_int(self.item_type)

    def serialize(self) -> dict[str, Any]:
        return {
            "item_id": self.item_id,
            "item_type": self.item_type,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> StorageItem:
        item = StorageItem(data.get("item_id", 0))
        item.item_type = data.get("item_type", 0)
        return item

    def __repr__(self) -> str:
        return f"StorageItem(item_id={self.item_id}, item_type={self.item_type})"

    def __str__(self) -> str:
        return f"StorageItem(item_id={self.item_id}, item_type={self.item_type})"


class Cats:
    def __init__(self, cats: list[Cat], total_storage_items: int = 0):
        self.cats = cats
        self.storage_items = [StorageItem.init() for _ in range(total_storage_items)]
        self.favourites: dict[int, bool] = {}
        self.chara_new_flags: dict[int, int] = {}
        self.unit_buy: UnitBuy | None = None
        self.unit_limit: UnitLimit | None = None
        self.nyanko_picture_book: NyankoPictureBook | None = None
        self.talent_data: TalentData | None = None

    def get_all_cats(self) -> list[Cat]:
        return self.cats

    @staticmethod
    def init(gv: core.GameVersion) -> Cats:
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            total_cats = 0
        cats_l: list[Cat] = []
        for i in range(total_cats):
            cats_l.append(Cat.init(i))

        if gv < 110100:
            total_storage_items = 100
        else:
            total_storage_items = 0
        return Cats(cats_l, total_storage_items)

    @staticmethod
    def get_gv_cats(gv: core.GameVersion) -> int | None:
        if gv == 20:
            total_cats = 203
        elif gv == 21:
            total_cats = 214
        elif gv == 22:
            total_cats = 231
        elif gv == 23:
            total_cats = 241
        elif gv == 24:
            total_cats = 249
        elif gv == 25:
            total_cats = 260
        else:
            total_cats = None
        return total_cats

    def get_unlocked_cats(self) -> list[Cat]:
        return [cat for cat in self.cats if cat.unlocked]

    def get_non_unlocked_cats(self) -> list[Cat]:
        return [cat for cat in self.cats if not cat.unlocked]

    def get_non_gacha_cats(self, save_file: core.SaveFile) -> list[Cat]:
        unitbuy = self.read_unitbuy(save_file)
        cats: list[Cat] = []
        for cat in self.cats:
            unit_buy_data = unitbuy.get_unit_buy(cat.id)
            if unit_buy_data is None:
                continue

            if unit_buy_data.unlock_source != 2:
                cats.append(cat)

        return cats

    def read_unitbuy(self, save_file: core.SaveFile) -> UnitBuy:
        if self.unit_buy is None:
            self.unit_buy = UnitBuy(save_file)
        return self.unit_buy

    def read_unitlimit(self, save_file: core.SaveFile) -> UnitLimit:
        if self.unit_limit is None:
            self.unit_limit = UnitLimit(save_file)
        return self.unit_limit

    def read_nyanko_picture_book(self, save_file: core.SaveFile) -> NyankoPictureBook:
        if self.nyanko_picture_book is None:
            self.nyanko_picture_book = NyankoPictureBook(save_file)
        return self.nyanko_picture_book

    def read_talent_data(self, save_file: core.SaveFile) -> TalentData | None:
        if self.talent_data is None:
            self.talent_data = TalentData.from_game_data(save_file)
        return self.talent_data

    def get_cats_rarity(self, save_file: core.SaveFile, rarity: int) -> list[Cat]:
        unit_buy = self.read_unitbuy(save_file)
        return [cat for cat in self.cats if unit_buy.get_cat_rarity(cat.id) == rarity]

    def get_cats_name(
        self,
        save_file: core.SaveFile,
        search_name: str,
    ) -> list[Cat]:
        cats: list[Cat] = []
        for cat in self.cats:
            names = cat.get_names_cls(save_file)
            if names is None:
                continue
            for name in names:
                if search_name.lower() in name.lower():
                    cats.append(cat)
                    break
        return cats

    def get_cats_obtainable(self, save_file: core.SaveFile) -> list[Cat] | None:
        nyanko_picture_book = self.read_nyanko_picture_book(save_file)
        obtainable_cats = nyanko_picture_book.get_obtainable_cats()
        if obtainable_cats is None:
            return None
        ny_cats = [cat.cat_id for cat in obtainable_cats]
        cats: list[Cat] = []
        for cat in self.cats:
            if cat.id in ny_cats:
                cats.append(cat)
        return cats

    def get_cats_non_obtainable(self, save_file: core.SaveFile) -> list[Cat] | None:
        nyanko_picture_book = self.read_nyanko_picture_book(save_file)
        obtainable_cats = nyanko_picture_book.get_obtainable_cats()
        if obtainable_cats is None:
            return None
        ny_cats = [cat.cat_id for cat in obtainable_cats]
        cats: list[Cat] = []
        for cat in self.cats:
            if cat.id not in ny_cats:
                cats.append(cat)
        return cats

    def get_cats_gatya_banner(
        self, save_file: core.SaveFile, gatya_id: int
    ) -> list[core.Cat] | None:
        cat_ids = save_file.gatya.read_gatya_data_set(save_file).get_cat_ids(gatya_id)
        if cat_ids is None:
            return None
        return self.get_cats_by_ids(cat_ids)

    def true_form_cats(
        self,
        save_file: core.SaveFile,
        cats: list[Cat],
        force: bool = False,
        set_current_forms: bool = True,
    ):
        pic_book = self.read_nyanko_picture_book(save_file)
        for cat in cats:
            pic_book_cat = pic_book.get_cat(cat.id)
            if force:
                cat.true_form(save_file, set_current_form=set_current_forms)
            elif pic_book_cat is not None:
                cat.set_form_true(
                    save_file,
                    pic_book_cat.total_forms,
                    set_current_form=set_current_forms,
                )

    def fourth_form_cats(
        self,
        save_file: core.SaveFile,
        cats: list[Cat],
        force: bool = False,
        set_current_forms: bool = True,
    ):
        pic_book = self.read_nyanko_picture_book(save_file)
        for cat in cats:
            pic_book_cat = pic_book.get_cat(cat.id)
            if force:
                cat.unlock_fourth_form(save_file, set_current_form=set_current_forms)
            elif pic_book_cat is not None:
                cat.set_form_true(
                    save_file,
                    pic_book_cat.total_forms,
                    set_current_form=set_current_forms,
                    fourth_form=True,
                )

    def get_cats_by_ids(self, ids: list[int]) -> list[Cat]:
        cats: list[Cat] = []
        for cat in self.cats:
            if cat.id in ids:
                cats.append(cat)
        return cats

    def get_cat_by_id(self, id: int) -> Cat | None:
        for cat in self.cats:
            if cat.id == id:
                return cat
        return None

    @staticmethod
    def get_rarity_names(save_file: core.SaveFile) -> list[str]:
        localizable = save_file.get_localizable()
        rarity_names: list[str] = []
        rarity_index = 1
        while True:
            rarity_name = localizable.get(f"rarity_name_{rarity_index}")
            if rarity_name is None:
                break
            rarity_names.append(rarity_name)
            rarity_index += 1
        return rarity_names

    @staticmethod
    def read_unlocked(stream: core.Data, gv: core.GameVersion) -> Cats:
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            total_cats = stream.read_int()
        cats_l: list[Cat] = []
        for i in range(total_cats):
            cats_l.append(Cat.read_unlocked(i, stream))
        return Cats(cats_l)

    def write_unlocked(self, stream: core.Data, gv: core.GameVersion):
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            stream.write_int(len(self.cats))
        for cat in self.cats:
            cat.write_unlocked(stream)

    def read_upgrade(self, stream: core.Data, gv: core.GameVersion):
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            total_cats = stream.read_int()
        for cat in self.cats:
            cat.read_upgrade(stream)

    def write_upgrade(self, stream: core.Data, gv: core.GameVersion):
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            stream.write_int(len(self.cats))
        for cat in self.cats:
            cat.write_upgrade(stream)

    def read_current_form(self, stream: core.Data, gv: core.GameVersion):
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            total_cats = stream.read_int()
        for cat in self.cats:
            cat.read_current_form(stream)

    def write_current_form(self, stream: core.Data, gv: core.GameVersion):
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            stream.write_int(len(self.cats))
        for cat in self.cats:
            cat.write_current_form(stream)

    def read_unlocked_forms(self, stream: core.Data, gv: core.GameVersion):
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            total_cats = stream.read_int()
        for cat in self.cats:
            cat.read_unlocked_forms(stream)

    def write_unlocked_forms(self, stream: core.Data, gv: core.GameVersion):
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            stream.write_int(len(self.cats))
        for cat in self.cats:
            cat.write_unlocked_forms(stream)

    def read_gatya_seen(self, stream: core.Data, gv: core.GameVersion):
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            total_cats = stream.read_int()
        for cat in self.cats:
            cat.read_gatya_seen(stream)

    def write_gatya_seen(self, stream: core.Data, gv: core.GameVersion):
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            stream.write_int(len(self.cats))
        for cat in self.cats:
            cat.write_gatya_seen(stream)

    def read_max_upgrade_levels(self, stream: core.Data, gv: core.GameVersion):
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            total_cats = stream.read_int()
        for cat in self.cats:
            cat.read_max_upgrade_level(stream)

    def write_max_upgrade_levels(self, stream: core.Data, gv: core.GameVersion):
        total_cats = Cats.get_gv_cats(gv)
        if total_cats is None:
            stream.write_int(len(self.cats))
        for cat in self.cats:
            cat.write_max_upgrade_level(stream)

    def read_storage(self, stream: core.Data, gv: core.GameVersion):
        if gv < 110100:
            total_storage = 100
        else:
            total_storage = stream.read_short()
        self.storage_items: list[StorageItem] = []
        for _ in range(total_storage):
            self.storage_items.append(StorageItem.read_item_id(stream))
        for item in self.storage_items:
            item.read_item_type(stream)

    def write_storage(self, stream: core.Data, gv: core.GameVersion):
        if gv >= 110100:
            stream.write_short(len(self.storage_items))
        for item in self.storage_items:
            item.write_item_id(stream)
        for item in self.storage_items:
            item.write_item_type(stream)

    def read_catguide_collected(self, stream: core.Data):
        total_cats = stream.read_int()
        for i in range(total_cats):
            self.cats[i].read_catguide_collected(stream)

    def write_catguide_collected(self, stream: core.Data):
        stream.write_int(len(self.cats))
        for cat in self.cats:
            cat.write_catguide_collected(stream)

    def read_fourth_forms(self, stream: core.Data):
        total_cats = stream.read_int()
        for i in range(total_cats):
            self.cats[i].read_fourth_form(stream)

    def read_catseyes_used(self, stream: core.Data):
        total_cats = stream.read_int()
        for i in range(total_cats):
            self.cats[i].read_catseyes_used(stream)

    def write_catseyes_used(self, stream: core.Data):
        stream.write_int(len(self.cats))
        for cat in self.cats:
            cat.write_catseyes_used(stream)

    def write_fourth_forms(self, stream: core.Data):
        stream.write_int(len(self.cats))
        for cat in self.cats:
            cat.write_fourth_form(stream)

    def read_favorites(self, stream: core.Data):
        self.favourites: dict[int, bool] = {}
        total_cats = stream.read_int()
        for _ in range(total_cats):
            cat_id = stream.read_int()
            self.favourites[cat_id] = stream.read_bool()

    def write_favorites(self, stream: core.Data):
        stream.write_int(len(self.favourites))
        for cat_id, is_favourite in self.favourites.items():
            stream.write_int(cat_id)
            stream.write_bool(is_favourite)

    def read_chara_new_flags(self, stream: core.Data):
        self.chara_new_flags: dict[int, int] = {}
        total_cats = stream.read_int()
        for _ in range(total_cats):
            cat_id = stream.read_int()
            self.chara_new_flags[cat_id] = stream.read_int()

    def write_chara_new_flags(self, stream: core.Data):
        stream.write_int(len(self.chara_new_flags))
        for cat_id, new_flag in self.chara_new_flags.items():
            stream.write_int(cat_id)
            stream.write_int(new_flag)

    def read_talents(self, stream: core.Data):
        total_cats = stream.read_int()
        for _ in range(total_cats):
            cat_id = stream.read_int()
            if cat_id < 0 or cat_id >= len(self.cats):
                cat = Cat.init(cat_id)
                cat.read_talents(stream)
                continue
            self.cats[cat_id].read_talents(stream)

    def write_talents(self, stream: core.Data):
        total_talents = 0
        for cat in self.cats:
            total_talents += 1 if cat.talents is not None else 0
        stream.write_int(total_talents)
        for cat in self.cats:
            if cat.talents is None:
                continue
            stream.write_int(cat.id)
            cat.write_talents(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "cats": [cat.serialize() for cat in self.cats],
            "storage_items": [item.serialize() for item in self.storage_items],
            "favorites": self.favourites,
            "chara_new_flags": self.chara_new_flags,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Cats:
        cats_l = [Cat.deserialize(cat) for cat in data.get("cats", [])]
        cats = Cats(cats_l)
        cats.storage_items = [
            StorageItem.deserialize(item) for item in data.get("storage_items", [])
        ]
        cats.favourites = data.get("favorites", {})
        cats.chara_new_flags = data.get("chara_new_flags", {})
        return cats

    def __repr__(self) -> str:
        return f"Cats(cats={self.cats}, storage_items={self.storage_items}, favourites={self.favourites}, chara_new_flags={self.chara_new_flags})"

    def __str__(self) -> str:
        return self.__repr__()



from __future__ import annotations

from bcsfe import core


class PowerUpHelper:
    def __init__(self, cat: core.Cat, save_file: core.SaveFile):
        self.cat = cat
        self.save_file = save_file
        self.unit_limit = self.save_file.cats.read_unitlimit(
            self.save_file
        ).get_unit_limit(self.cat.id)
        self.all_unit_buy = self.save_file.cats.read_unitbuy(self.save_file)
        self.unit_buy = self.all_unit_buy.get_unit_buy(self.cat.id)
        self.rank_gifts = self.save_file.user_rank_rewards.read_rank_gifts(
            self.save_file
        )
        self.max_upgrade_level = self.__get_max_upgrade_level_check()

    def get_current_max_level(self) -> int | None:
        if self.unit_buy is None:
            return None
        return min(
            self.unit_buy.original_max_levels[0] + self.max_upgrade_level,
            self.unit_buy.max_upgrade_level_catseye,
        )

    def has_strict_upgrade(self) -> bool:
        return core.core_data.config.get_bool(core.ConfigKey.STRICT_UPGRADE)

    def get_upgrade_state_check(self) -> int:
        if not self.has_strict_upgrade():
            return 100000
        return self.save_file.upgrade_state

    def get_user_rank_check(self) -> int:
        if not self.has_strict_upgrade():
            return 1000000
        return self.save_file.calculate_user_rank()

    def __get_max_upgrade_level_check(self) -> int:
        if self.unit_limit is None:
            return self.cat.max_upgrade_level.base

        rewards = self.save_file.user_rank_rewards
        self.cat.max_upgrade_level.reset()

        strict_upgrade = self.has_strict_upgrade()

        for reward_id in range(len(rewards.rewards)):
            rank_gift = self.rank_gifts.get_by_id(reward_id)
            if rank_gift is None:
                continue
            user_rank_reward = rewards.rewards[reward_id]
            if not user_rank_reward.claimed and strict_upgrade:
                continue
            for present in rank_gift.rewards:
                if present[0] >= 1000 and present[0] <= 1599:
                    for limit in self.unit_limit.values:
                        if limit == present[0]:
                            self.cat.max_upgrade_level.increment_base(
                                present[1]
                            )
                elif present[0] >= 4000 and present[0] <= 4599:
                    for limit in self.unit_limit.values:
                        if limit == present[0]:
                            self.cat.max_upgrade_level.increment_plus(
                                present[1]
                            )

        return self.cat.max_upgrade_level.base

    def can_power_up(self) -> bool:
        if self.unit_buy is None:
            return False
        base_level = self.cat.upgrade.get_base()
        current_max_level = self.get_current_max_level()
        if current_max_level is None:
            return False

        if base_level >= current_max_level or (
            (
                self.get_upgrade_state_check() > 1
                or base_level == self.unit_buy.unknown_22
            )
            and self.get_upgrade_state_check() < 2
        ):
            return (
                self.unit_buy.rarity != 0
                and base_level >= self.unit_buy.max_upgrade_level_no_catseye
                and base_level < self.unit_buy.max_upgrade_level_catseye
                and base_level < current_max_level
            )
        return True

    def can_use_catseye(self) -> bool:
        if self.unit_buy is None:
            return False

        base_level = self.cat.upgrade.get_base()
        return (
            self.unit_buy.rarity != 0
            and base_level >= self.unit_buy.max_upgrade_level_no_catseye
            and self.unit_buy.max_upgrade_level_no_catseye != -1
            and self.get_user_rank_check() >= 1600
        )

    def upgrade_cat(self, force: bool = False) -> bool:
        if force:
            self.cat.upgrade_base(self.save_file)
            return True
        if self.unit_buy is None:
            return False
        current_max_level = self.get_current_max_level()
        if current_max_level is None:
            return False

        if self.can_power_up():
            self.cat.upgrade_base(self.save_file)
            return True

        if (
            self.can_use_catseye()
            and self.unit_buy.max_upgrade_level_no_catseye <= current_max_level
        ):
            if (
                self.cat.upgrade.get_base()
                < self.unit_buy.max_upgrade_level_catseye
            ):
                self.cat.upgrade_base(self.save_file)
                self.cat.catseyes_used += 1
                self.cat.max_upgrade_level.upgrade()
                return True
            return False
        return False

    def get_max_max_base_upgrade_level(self) -> int:
        max_level = 0
        if self.all_unit_buy.unit_buy is None:
            return 90
        for unit_buy in self.all_unit_buy.unit_buy:
            if unit_buy.max_upgrade_level_catseye > max_level:
                max_level = unit_buy.max_upgrade_level_catseye
        return max_level

    def get_max_max_plus_upgrade_level(self) -> int:
        max_level = 0
        if self.all_unit_buy.unit_buy is None:
            return 90
        for unit_buy in self.all_unit_buy.unit_buy:
            if unit_buy.max_plus_upgrade_level > max_level:
                max_level = unit_buy.max_plus_upgrade_level
        return max_level

    def get_max_possible_base(self) -> int:
        if self.unit_buy is None:
            return 90
        return self.unit_buy.max_upgrade_level_catseye

    def get_max_possible_plus(self) -> int:
        if self.unit_buy is None:
            return 90
        return self.unit_buy.max_plus_upgrade_level

    def reset_upgrade(self):
        self.cat.upgrade.base = 0
        self.cat.catseyes_used = 0

    def upgrade_by(self, amount: int):
        if amount == -1:
            return
        for _ in range(amount):
            did_upgrade = self.upgrade_cat()
            if not did_upgrade:
                break

    def max_upgrade(self):
        while self.upgrade_cat():
            pass
